from datetime import timedelta, datetime

import allure
import pytest

import refinitiv.data as rd
from tests.integration.access.conftest import (
    check_column_names_is_exist_in_response_and_df_not_empty,
    check_rics_fields_order_respects_and_df_not_empty,
    check_df_index_names,
    check_df_contains_ohlc_columns_and_universes,
)
from tests.integration.helpers import (
    check_response_data_start_end_date,
    check_index_column_contains_dates,
    check_dataframe_column_date_for_datetime_type,
    check_if_dataframe_is_not_none,
    check_the_number_of_items_in_dataframe,
    check_universe_order_in_df,
)


@allure.suite("FinCoder layer")
@allure.feature("FinCoder layer - get_history")
@allure.severity(allure.severity_level.CRITICAL)
@pytest.mark.underlying_content("datagrid")
class TestGetHistory:
    @allure.title("Getting history for single instrument")
    @pytest.mark.parametrize(
        "universe,fields,expected_fields",
        [
            (["IBM.N", "EUR="], None, None),
            ("LSEG.L", None, None),
            ("LSEG.L", "TR.Revenue", "Revenue"),
        ],
    )
    @pytest.mark.caseid("C37270766_1")
    @pytest.mark.smoke
    def test_get_history_for_single_instrument_and_all_fields(
        self,
        set_underlying_platform_config,
        open_desktop_session,
        universe,
        fields,
        expected_fields,
    ):
        """
        result1 = rd.get_history(
            universe=["IBM.N", "EUR="], fields=None
        )
        result2 = rd.get_history(
            universe="LSEG.L", fields=None
        )
        result3 = rd.get_history(
            universe="LSEG.L", fields="TR.Revenue"
        )
        """
        response = rd.get_history(
            universe=universe, fields=fields, use_field_names_in_headers=False
        )
        check_column_names_is_exist_in_response_and_df_not_empty(
            response, expected_fields
        )
        check_index_column_contains_dates(response)
        check_if_dataframe_is_not_none(response)

    @allure.title("Getting history for multiple instruments with commingle fields")
    @pytest.mark.parametrize(
        "universe,fields,interval,start_date,end_date,expected_fields",
        [
            (
                "IBM",
                [
                    "BID",
                    "TR.RevenueMean.currency",
                    "TR.RevenueMean",
                    "TR.RevenueMean",
                ],
                "1h",
                timedelta(-10),
                timedelta(0),
                [
                    "BID",
                    "Currency",
                    "Revenue - Mean",
                ],
            ),
            (
                (
                    ["IBM", "VOD.L"],
                    [
                        "BID",
                        "TR.RevenueMean.currency",
                        "TR.RevenueMean",
                    ],
                    "hourly",
                    "2022-06-01T11:00:00",
                    "2022-06-06T11:00:00",
                    [
                        ("IBM", "BID"),
                        ("IBM", "Currency"),
                        ("IBM", "Revenue - Mean"),
                        ("VOD.L", "BID"),
                        ("VOD.L", "Currency"),
                        ("VOD.L", "Revenue - Mean"),
                    ],
                )
            ),
        ],
        ids=["timedelta", "string"],
    )
    @pytest.mark.caseid("37270766")
    @pytest.mark.smoke
    def test_get_history_for_multiple_instruments(
        self,
        set_underlying_platform_config,
        open_desktop_session,
        universe,
        fields,
        interval,
        start_date,
        end_date,
        expected_fields,
        request,
    ):
        """
        response1 = rd.get_history(
            universe="IBM",
            fields=[
                "BID",
                "TR.RevenueMean.currency",
                "TR.RevenueMean",
                "TR.RevenueMean",
            ],
            interval="1h",
            start=timedelta(-10),
            end=timedelta(0),
        )

        response2 = rd.get_history(
            universe=["IBM", "VOD.L"],
            fields=[
                "BID",
                "TR.RevenueMean.currency",
                "TR.RevenueMean",
            ],
            interval="hourly",
            start="2022-06-01T11:00:00",
            end="2022-06-06T11:00:00",
        )

        """
        response = rd.get_history(
            universe=universe,
            fields=fields,
            interval=interval,
            start=start_date,
            end=end_date,
        )

        check_df_index_names(response, interval)
        check_column_names_is_exist_in_response_and_df_not_empty(
            response, expected_fields
        )
        check_index_column_contains_dates(response)
        check_response_data_start_end_date(response, start_date, end_date, request)

    @allure.title("Getting history for instruments with platform session")
    @pytest.mark.parametrize(
        "universe,fields,interval,start_date,end_date,expected_fields",
        [
            (
                ["EUR=", "VOD.L"],
                [
                    "BID",
                    "TR.RevenueMean.currency",
                    "TR.RevenueMean",
                    "TR.RevenueMean",
                ],
                "daily",
                timedelta(-30),
                timedelta(0),
                [
                    ("EUR=", "BID"),
                    ("EUR=", "Currency"),
                    ("EUR=", "Revenue - Mean"),
                    ("VOD.L", "BID"),
                    ("VOD.L", "Currency"),
                    ("VOD.L", "Revenue - Mean"),
                ],
            )
        ],
        ids=["timedelta"],
    )
    @pytest.mark.caseid("C37270766_2")
    def test_get_history_for_instruments_with_platform_session(
        self,
        set_underlying_platform_config,
        open_platform_session_with_rdp_creds,
        universe,
        fields,
        interval,
        start_date,
        end_date,
        expected_fields,
        request,
    ):
        response = rd.get_history(
            universe=universe,
            fields=fields,
            interval=interval,
            start=start_date,
            end=end_date,
        )
        check_df_index_names(response, interval)
        check_column_names_is_exist_in_response_and_df_not_empty(
            response, expected_fields
        )
        check_index_column_contains_dates(response)
        check_response_data_start_end_date(response, start_date, end_date, request)

    @allure.title("Getting history with duplicated instruments")
    @pytest.mark.parametrize(
        "universes,fields,interval,start_date,end_date,expected_fields",
        [
            (
                (
                    ["IBM", "IBM", "VOD.L"],
                    [
                        "ASK",
                        "TR.RevenueMean.currency",
                        "TR.RevenueMean",
                    ],
                    "monthly",
                    datetime(2022, 3, 10),
                    datetime(2022, 5, 31),
                    [
                        ("IBM", "ASK"),
                        ("IBM", "Currency"),
                        ("IBM", "Revenue - Mean"),
                        ("IBM", "ASK"),
                        ("IBM", "Currency"),
                        ("IBM", "Revenue - Mean"),
                        ("VOD.L", "ASK"),
                        ("VOD.L", "Currency"),
                        ("VOD.L", "Revenue - Mean"),
                    ],
                )
            ),
        ],
        ids=["datetime"],
    )
    @pytest.mark.caseid("37270767")
    def test_get_history_with_duplicated_instruments(
        self,
        set_underlying_platform_config,
        open_desktop_session,
        universes,
        fields,
        interval,
        start_date,
        end_date,
        expected_fields,
        request,
    ):
        response = rd.get_history(
            universe=universes,
            fields=fields,
            interval=interval,
            start=start_date,
            end=end_date,
        )
        list_of_instruments = list(response.columns)
        check_the_number_of_items_in_dataframe(
            expected_fields[0], list_of_instruments, universes.count(universes[0])
        )
        check_index_column_contains_dates(response)
        check_response_data_start_end_date(response, start_date, end_date, request)
        check_df_index_names(response, interval)

    @allure.title("Getting history with invalid instrument")
    @pytest.mark.parametrize(
        "universe,fields,interval,start_date,end_date,expected_fields",
        [
            (
                (
                    ["IBM", "INVALID", "VOD.L"],
                    [
                        "BID",
                        "TR.RevenueMean",
                    ],
                    "1Y",
                    datetime(2018, 3, 10),
                    datetime(2021, 12, 31),
                    [
                        ("IBM", "BID"),
                        ("IBM", "Revenue - Mean"),
                        ("VOD.L", "BID"),
                        ("VOD.L", "Revenue - Mean"),
                        ("INVALID", "BID"),
                        ("INVALID", "Revenue - Mean"),
                    ],
                )
            )
        ],
        ids=["datetime"],
    )
    @pytest.mark.caseid("37270776")
    def test_get_history_with_invalid_instruments(
        self,
        set_underlying_platform_config,
        open_desktop_session,
        universe,
        fields,
        interval,
        start_date,
        end_date,
        expected_fields,
        request,
    ):
        response = rd.get_history(
            universe=universe,
            fields=fields,
            interval=interval,
            end=end_date,
            start=start_date,
        )

        check_column_names_is_exist_in_response_and_df_not_empty(
            response, expected_fields
        )
        check_index_column_contains_dates(response)
        check_response_data_start_end_date(response, start_date, end_date, request)
        check_df_index_names(response, interval)

    @allure.title("Getting pricing history for single and multiple instruments")
    @pytest.mark.parametrize(
        "universe,fields,interval,expected_fields",
        [
            (
                ["EUR="],
                ["BID", "ASK"],
                "daily",
                ["BID", "ASK"],
            ),
            (
                ["EUR=", "GBP="],
                ["BID", "ASK"],
                "daily",
                [("GBP=", "BID"), ("GBP=", "ASK"), ("EUR=", "BID"), ("EUR=", "ASK")],
            ),
            (
                ["EUR=", "GBP="],
                ["BID", "ASK"],
                "tick",
                [
                    ("GBP=", "BID"),
                    ("GBP=", "ASK"),
                    ("EUR=", "BID"),
                    ("EUR=", "ASK"),
                ],
            ),
        ],
    )
    @pytest.mark.caseid("37270777")
    def test_get_pricing_history_for_single_and_multiple_instruments(
        self,
        set_underlying_platform_config,
        open_desktop_session,
        universe,
        fields,
        interval,
        expected_fields,
    ):
        response = rd.get_history(
            universe=universe,
            fields=fields,
            interval=interval,
            start=datetime(2022, 5, 24),
        )

        check_column_names_is_exist_in_response_and_df_not_empty(
            response, expected_fields
        )
        check_index_column_contains_dates(response)
        check_df_index_names(response, interval)

    @allure.title("Getting history for pricing fields")
    @pytest.mark.parametrize(
        "universe,fields,interval,end",
        [
            (
                "IBM",
                ["HIGH_1", "LOW_1", "TRDPRC_1", "NUM_MOVES"],
                "daily",
                "2020-01-01",
            ),
            (
                ["LSEG.L", "VOD.L", "GOOG.O", "AAPL.OQ", "IBM.N"],
                ["ASK", "TRNOVR_UNS", "VWAP", "BLKCOUNT", "BLKVOLUM"],
                "1Y",
                "2022-01-01",
            ),
        ],
        ids=["single", "multiple"],
    )
    @pytest.mark.caseid("38386444")
    def test_get_history_pricing_fields(
        self,
        set_underlying_platform_config,
        open_desktop_session,
        request,
        universe,
        fields,
        interval,
        end,
    ):
        response = rd.get_history(
            universe=universe, fields=fields, interval=interval, end=end
        )

        if "single" in request.node.callspec.id:
            check_column_names_is_exist_in_response_and_df_not_empty(response, fields)
        else:
            check_rics_fields_order_respects_and_df_not_empty(
                response, universe, fields
            )
        check_index_column_contains_dates(response)
        check_df_index_names(response, interval)

    @allure.title("Getting history for fundamental and reference fields")
    @pytest.mark.parametrize(
        "universe,fields,interval,use_field_names_in_headers,expected_fields_udf,expected_fields_rdp",
        [
            (
                "IBM",
                ["TR.Revenue.date", "TR.RevenueMean.currency", "TR.RevenueMean"],
                "daily",
                True,
                ["TR.REVENUEMEAN", "TR.REVENUE.DATE", "TR.REVENUEMEAN.currency"],
                ["TR.Revenue", "TR.RevenueMean"],
            ),
            (
                "GOOG.O",
                ["TR.Revenue.date", "TR.Revenue"],
                "daily",
                False,
                ["Date", "Revenue"],
                ["Date", "Revenue"],
            ),
            (
                ["LSEG", "GOOG.O"],
                "TR.Revenue",
                "daily",
                False,
                ["LSEG", "GOOG.O"],
                ["LSEG", "GOOG.O"],
            ),
        ],
        ids=["single_field_names", "single_field_titles", "multiple_title_names"],
    )
    @pytest.mark.caseid("38386491")
    def test_get_history_fundamental_and_reference_fields(
        self,
        set_underlying_platform_config,
        open_desktop_session,
        universe,
        fields,
        interval,
        use_field_names_in_headers,
        expected_fields_rdp,
        expected_fields_udf,
    ):
        response = rd.get_history(
            universe=universe,
            fields=fields,
            interval=interval,
            use_field_names_in_headers=use_field_names_in_headers,
        )
        expected_fields = expected_fields_udf
        if set_underlying_platform_config == "rdp-underlying-platform":
            expected_fields = expected_fields_rdp
        check_column_names_is_exist_in_response_and_df_not_empty(
            response, expected_fields
        )
        check_dataframe_column_date_for_datetime_type(response)

    @allure.title("Getting mix of pricing and fundamental history")
    @pytest.mark.parametrize(
        "universe,fields,interval,expected_fields",
        [
            (
                ["IBM", "EUR="],
                [
                    "BID",
                    "TR.RevenueMean.currency",
                ],
                "daily",
                [
                    ("EUR=", "BID"),
                    ("EUR=", "Currency"),
                    ("IBM", "BID"),
                    ("IBM", "Currency"),
                ],
            ),
            (
                (
                    ["VOD.L", "EUR="],
                    [
                        "BID",
                        "ASK",
                        "EVENT_TYPE",
                    ],
                    "tick",
                    [
                        ("EUR=", "EVENT_TYPE"),
                        ("EUR=", "BID"),
                        ("EUR=", "ASK"),
                        ("VOD.L", "EVENT_TYPE"),
                        ("VOD.L", "BID"),
                        ("VOD.L", "ASK"),
                    ],
                )
            ),
        ],
    )
    @pytest.mark.caseid("37270778")
    def test_get_mix_of_pricing_and_fundamental_history(
        self,
        set_underlying_platform_config,
        open_desktop_session,
        universe,
        fields,
        interval,
        expected_fields,
    ):
        response = rd.get_history(
            universe=universe,
            fields=fields,
            interval=interval,
        )

        check_column_names_is_exist_in_response_and_df_not_empty(
            response, expected_fields
        )
        check_index_column_contains_dates(response)
        check_df_index_names(response, interval)

    @allure.title("Getting history raises error on invalid fields list")
    @pytest.mark.parametrize(
        "universe,fields,interval",
        [
            (
                "IBM",
                ["INVALID_FIELD_1", "INVALID_FIELD_2"],
                "daily",
            ),
            (
                "INVAL",
                ["TR.Revenue.date", "TR.Revenue", "TR.GrossProfit"],
                "tick",
            ),
        ],
    )
    @pytest.mark.caseid("38386507")
    def test_get_history_with_invalid_fields(
        self,
        set_underlying_platform_config,
        open_desktop_session,
        universe,
        fields,
        interval,
    ):
        response = rd.get_history(universe=universe, fields=fields, interval=interval)

        assert response.empty, f"Response is empty: {response}"

    @allure.title("Getting history with resampling ohlc data")
    @pytest.mark.caseid("C43506364")
    @pytest.mark.parametrize(
        "universe,fields,interval,count",
        [
            (
                ["VOD.L", "GOOG.O"],
                ["BID", "ASK"],
                "tick",
                50,
            )
        ],
    )
    def test_get_history_with_bars_resampling(
        self,
        open_desktop_session,
        universe,
        fields,
        interval,
        count,
    ):
        response = rd.get_history(
            universe=universe, fields=fields, interval=interval, count=count
        )
        assert response.size == count * 2 * len(universe) * len(
            fields
        ), f"Inconsistency with the amount of returned data "
        dataframe = response.ohlc("2s")
        assert not response.empty, "Empty dataframe received"
        check_index_column_contains_dates(dataframe)
        check_df_contains_ohlc_columns_and_universes(dataframe, fields, universe)
        check_df_index_names(dataframe)
        check_universe_order_in_df(response, universe)

    @allure.title("Getting history by function and chain")
    @pytest.mark.parametrize(
        "universe,fields,expected_universe",
        [
            (
                'SCREEN(U(IN(Equity(active,public,primary))/*UNV:Public*/), IN(TR.HQCountryCode,"AR"), IN(TR.GICSIndustryCode,"401010"))',
                [
                    "TR.Revenue",
                    "BID",
                ],
                None,
            ),
            ('Peers("VOD.L")', ["BID", "TR.Revenue"], [("LBTYA.OQ", "BID")]),
            pytest.param(
                ["AAPL.OQ", "Peers('IBM')", "0#.DJI"],
                ["BID", "TR.Revenue"],
                [("AAPL.OQ", "BID"), ("CRM.N", "BID")],
                id="chain",
            ),
            (
                "GOOG.O",
                [
                    "ASK",
                    "BID",
                    "AVAIL(TR.GrossProfit(Period=LTM,Methodology=InterimSum)/TR.TotalRevenue(Period=LTM,Methodology=InterimSum),TR.GrossProfit(Period=FY0)/TR.TotalRevenue(Period=FY0))*100",
                ],
                ["ASK", "BID"],
            ),
        ],
    )
    @pytest.mark.caseid("37270766")
    @pytest.mark.smoke
    def test_get_history_by_function_and_chain(
        self,
        request,
        set_underlying_platform_config,
        open_desktop_session,
        universe,
        fields,
        expected_universe,
    ):
        response = rd.get_history(
            universe=universe,
            fields=fields,
        )

        check_index_column_contains_dates(response)
        check_if_dataframe_is_not_none(response)

        list_of_instruments = list(response.columns)
        if expected_universe:
            for ric in expected_universe:
                assert ric in list_of_instruments

        if "chain" in request.node.callspec.id:
            check_the_number_of_items_in_dataframe(
                expected_universe[0], list_of_instruments, len(universe)
            )

    @allure.title("Getting CustomInstrument data")
    @pytest.mark.parametrize(
        "universe",
        [([]), (["IBM", "EUR="])],
        ids=["ci_instruments", "mix_universes"],
    )
    @pytest.mark.caseid("C43973495_1")
    def test_get_custom_instrument_data(
        self,
        set_underlying_platform_config,
        open_desktop_session,
        create_instrument,
        request,
        universe,
    ):
        symbol_01 = create_instrument()
        symbol_02 = create_instrument()
        universes = universe + [symbol_01, symbol_02]

        if "ci_instruments" in request.node.callspec.id:
            expected_fields = [
                (symbol_01, "TRDPRC_1"),
                (symbol_02, "TRDPRC_1"),
            ]
        else:
            expected_fields = None

        response = rd.get_history(universe=universes)

        check_column_names_is_exist_in_response_and_df_not_empty(
            response, expected_fields
        )
        check_index_column_contains_dates(response)
        check_if_dataframe_is_not_none(response)
        check_df_index_names(response, interval="daily")

    @allure.title("Getting mix of pricing, fundamental and CustomInstrument history")
    @pytest.mark.parametrize(
        "universe,fields,interval,start_date,end_date,expected_fields",
        [
            (
                ["IBM", "EUR="],
                ["TR.Revenue", "TR.PriceClose", "TRDPRC_1"],
                "1M",
                datetime.strptime("01-01-2022 00:00:00", "%d-%m-%Y %H:%M:%S"),
                datetime.strptime("01-07-2022 00:00:00", "%d-%m-%Y %H:%M:%S"),
                [
                    ("IBM", "Revenue"),
                    ("IBM", "Price Close"),
                    ("IBM", "TRDPRC_1"),
                    ("EUR=", "Revenue"),
                    ("EUR=", "Price Close"),
                    ("EUR=", "TRDPRC_1"),
                ],
            ),
        ],
    )
    @pytest.mark.caseid("C43973495_2")
    def test_get_mix_of_pricing_custom_instrument_fundamental_history(
        self,
        set_underlying_platform_config,
        open_desktop_session,
        create_instrument,
        universe,
        fields,
        interval,
        start_date,
        end_date,
        expected_fields,
    ):
        symbol_01 = create_instrument()
        symbol_02 = create_instrument()
        universes = universe + [symbol_01, symbol_02]
        expected_data = expected_fields + [
            (symbol_01, "TRDPRC_1"),
            (symbol_01, "Price Close"),
            (symbol_01, "Revenue"),
            (symbol_02, "TRDPRC_1"),
            (symbol_02, "Price Close"),
            (symbol_02, "Revenue"),
        ]

        response = rd.get_history(
            universe=universes,
            fields=fields,
            interval=interval,
            start=start_date,
            end=end_date,
        )

        check_column_names_is_exist_in_response_and_df_not_empty(
            response, expected_data
        )
        check_index_column_contains_dates(response)
        check_df_index_names(response, interval)

    @allure.title("Getting history by function and chain without ADC fields")
    @pytest.mark.parametrize(
        "universe,fields",
        [
            ("0#.FCHI", ["BID", "ASK"]),
            ('Peers("VOD.L")', None),
            (
                'SCREEN(U(IN(Equity(active,public,primary))/*UNV:Public*/), IN(TR.HQCountryCode,"AR"), IN(TR.GICSIndustryCode,"401010"))',
                [
                    "ASK" "BID",
                ],
            ),
        ],
    )
    @pytest.mark.caseid("C54271588")
    def test_get_history_by_function_and_chain_without_adc_fields(
        self, open_session, universe, fields
    ):
        response = rd.get_history(
            universe=universe,
            fields=fields,
        )

        assert response.empty, f"Dataframe is not empty"



import time
from typing import TYPE_CHECKING, Callable

import requests

from .event import UpdateEvent
from .tools import NullResponse, UNAUTHORIZED_CODES
from .updater import Updater
from ..log_reporter import LogReporter
from ...delivery._data._endpoint_data import RequestMethod
from ...delivery._data._request import Request

if TYPE_CHECKING:
    from ._platform_session import PlatformSession
    from .grant_password import GrantPassword

codes = requests.codes


class AccessTokenUpdater(Updater, LogReporter):
    def __init__(
        self,
        session: "PlatformSession",
        delay: int,
        callback: Callable[[str, str, dict], None],
    ):
        Updater.__init__(self, delay, "AccessTokenUpdater")
        LogReporter.__init__(self, logger=session)

        self._session = session
        self._callback = callback

        self._grant: "GrantPassword" = self._session._grant
        self._app_key: str = self._session._app_key
        self._url: str = self._session.authentication_token_endpoint_url
        self._signon_control: bool = self._session.signon_control

        self.latency_secs: float = 0.0

    @Updater.delay.setter
    def delay(self, value: int):
        if value <= 0:
            raise ValueError("Delay must be greater than 0")
        Updater.delay.fset(self, value)

    def _do_update(self):
        response = self._request_token(
            self._grant, self._app_key, self._url, self._signon_control
        )
        status_code = response.status_code
        json_content = response.json()

        if status_code == codes.ok:
            event = UpdateEvent.ACCESS_TOKEN_SUCCESS
            message = "All is well"

        elif status_code in UNAUTHORIZED_CODES:
            event = UpdateEvent.ACCESS_TOKEN_UNAUTHORIZED
            error = json_content.get("error", "empty error")
            error_description = json_content.get(
                "error_description", "empty error description"
            )
            message = error_description
            self.error(f"[Error {status_code} - {error}] {error_description}")

        else:
            event = UpdateEvent.ACCESS_TOKEN_FAILED
            error = json_content.get("error")
            error_description = json_content.get(
                "error_description",
                getattr(response, "text", "empty error description"),
            )
            message = error_description
            self.error(f"[Error {status_code} - {error}] {error_description}")

        self._callback(event, message, json_content)

    def _request_token(
        self, grant: "GrantPassword", app_key: str, url: str, take_signon_control: bool
    ):
        username = grant.get_username()
        data = {
            "scope": grant.get_token_scope(),
            "grant_type": "password",
            "username": username,
            "password": grant.get_password(),
            "takeExclusiveSignOnControl": "true" if take_signon_control else "false",
        }

        if app_key is not None:
            data["client_id"] = app_key

        headers = {"Accept": "application/json"}
        try:
            start = time.time()
            request = Request(
                url=url, method=RequestMethod.POST, headers=headers, data=data
            )
            response = self._session.http_request(request)
            end = time.time()
            self.latency_secs = end - start
            self.debug(
                f"Latency: {self.latency_secs} sec.\n"
                f"Access token response: {response.text}"
            )
        except Exception as e:
            response = NullResponse()
            response.text = str(e)

        return response

    def _do_dispose(self):
        self._session = None
        self._callback = None



import enum
import re
import traceback
from collections.abc import Callable
from typing import TYPE_CHECKING, Optional

from ..._content_type import ContentType
from ..._tools import cached_property, make_enum_arg_parser
from ...delivery._stream import StreamStateEvent, StreamStateManager
from ...delivery._stream._stream_factory import create_rdp_stream

if TYPE_CHECKING:
    from ..._types import ExtendedParams, Strings, OptStr, OptStrStrs, OptCall
    from ..._core.session import Session
    from ...delivery._stream import _RDPStream

QUEUE_SIZE_PATTERN = re.compile(r"^queueSize=(?P<queue_size>[0-9]+)")


class Events(enum.Enum):
    """Events"""

    No = "None"
    Full = "Full"


class FinalizedOrders(enum.Enum):
    """Finalized order in cached"""

    No = "None"
    P1D = "P1D"


class UniverseTypes(enum.Enum):
    """Universe Types"""

    RIC = "RIC"
    Symbol = "Symbol"
    UserID = "UserID"


universe_type_arg_parser = make_enum_arg_parser(UniverseTypes, can_be_lower=True)
finalized_orders_arg_parser = make_enum_arg_parser(FinalizedOrders, can_be_lower=True)
events_arg_parser = make_enum_arg_parser(Events, can_be_lower=True)


class TradeDataStream(StreamStateManager):
    """
    Open a streaming trading analytics subscription.

    Parameters
    ----------
    universe: list
        a list of RIC or symbol or user's id for retrieving trading analytics data.

    fields: list
        a list of enumerate fields.
        Default: None

    universe_type: enum
        a type of given universe can be RIC, Symbol or UserID.
        Default: UniverseTypes.RIC

    finalized_orders: bool
        enable/disable the cached of finalized order of current day in the streaming.
        Default: False

    filters: list
        set the condition of subset of trading streaming data
        Default: None

    """

    def __init__(
        self,
        session: "Session",
        universe: "OptStrStrs" = None,
        universe_type: Optional["UniverseTypes"] = None,
        fields: "OptStrStrs" = None,
        events: Optional["Events"] = None,
        finalized_orders: Optional["FinalizedOrders"] = None,
        filters: "OptStrStrs" = None,
        api: "OptStr" = None,
        extended_params: "ExtendedParams" = None,
    ):
        self._session = session

        StreamStateManager.__init__(self, logger=self._session.logger())

        self._universe = universe
        self._universe_type = universe_type
        self._fields = fields
        self._event_details = events
        self._finalized_orders = finalized_orders
        self._filters = filters
        self._api = api
        self._extended_params = extended_params

        self.on_update: "OptCall" = None
        self.on_complete: "OptCall" = None
        self.on_add: "OptCall" = None
        self.on_remove: "OptCall" = None
        self.on_event: "OptCall" = None
        self.on_state: "OptCall" = None

        self._headers_ids: "Strings" = []

        self._is_completed: bool = False

    @property
    def session(self) -> "Session":
        return self._session

    @session.setter
    def session(self, session: "Session"):
        if self._session != session and not self.is_open:
            self._session = session
            StreamStateManager.init_logger(self, self._session.logger())
            self._stream.session = self._session

    @cached_property
    def _stream(self) -> "_RDPStream":
        parameters = {
            "universeType": self._universe_type,
            "events": self._event_details,
            "finalizedOrders": self._finalized_orders,
        }
        if self._filters is not None:
            parameters["filters"] = self._filters

        view = None
        if self._fields:
            view = self._fields.copy()

        content_type = ContentType.STREAMING_TRADING
        if self._api:
            content_type = ContentType.STREAMING_CUSTOM

        stream = create_rdp_stream(
            content_type=content_type,
            session=self._session,
            universe=self._universe,
            view=view,
            parameters=parameters,
            api=self._api,
            extended_params=self._extended_params,
        )
        stream.on_response(self._do_on_response)
        stream.on_update(self._do_on_update)
        stream.on(StreamStateEvent.CLOSED, self.close)
        return stream

    def _do_open(self, *args, **kwargs):
        self._debug(f"{self._classname} Open stream for {self._universe}")
        self._stream.open()

    def _do_close(self, *args, **kwargs):
        self._debug(f"{self._classname} Close stream for {self._universe}")
        self._stream.close()

    def _do_on_response(self, stream: "_RDPStream", message: dict):
        """
        Extract the response order summaries, order events and state

        Parameters
        ----------
        stream: _RDPStream
        message: dict
            {
                'streamID': '5',
                'type': 'Response',
                'headers': [
                            {'id': 'OrderKey', 'type': 'String'},
                            {'id': 'OrderTime', 'type': 'String', 'format': 'datetime'},
                            {'id': 'OrderStatus', 'type': 'String'},
                ],
                'state': {
                    'code': 200,
                    'status': 'Ok',
                    'stream': 'Open',
                    'message': 'queueSize=133'
                }
            }

        Returns
        -------
        None
        """
        self._headers_ids = [hdr["id"] for hdr in message.get("headers", [])]

        self._process_data(message)

        messages_data = message.get("messages", [])
        for datum in messages_data:
            self._callback(self.on_event, datum)

        self._process_state(message)

    def _do_on_update(self, stream: "_RDPStream", message: dict):
        """
        Extract the update (add/update/remove) order summaries and new order status.

        Parameters
        ----------
        stream: _RDPStream
        message: dict

        Returns
        -------
        None
        """
        self._process_data(message)

        update_data = message.get("update", [])
        for datum in update_data:
            self._callback(self.on_update, datum)

        removed_data = message.get("remove", [])
        for datum in removed_data:
            self._callback(self.on_remove, datum)

        messages_data = message.get("messages", [])
        for datum in messages_data:
            self._callback(self.on_event, datum)

        self._process_state(message)

    def _process_data(self, message: dict) -> None:
        data = message.get("data", [])
        for datum in data:
            self._callback(self.on_add, dict(zip(self._headers_ids, datum)))

    def _process_state(self, message: dict) -> None:
        state = message.get("state", {})

        if "message" in state:
            matched = QUEUE_SIZE_PATTERN.match(state["message"])

            if matched is not None:
                group = matched.groupdict()
                queue_size = group.get("queue_size", -1)
                queue_size = int(queue_size)

                if queue_size == 0 and not self._is_completed:
                    self._is_completed = True
                    self._callback(self.on_complete)

        if state:
            self._callback(self.on_state, state)

    def _callback(self, func: Callable, message: dict = None):
        if func:
            try:
                if message:
                    func(self, message)
                else:
                    func(self)
            except Exception as e:
                self._error(f"{self._classname} {func} callback raised exception:{e!r}")
                self._debug(f"{traceback.format_exc()}")



import itertools
from unittest import mock
from unittest.mock import patch

import pytest

from refinitiv.data._content_type import ContentType
from refinitiv.data.content._universe_stream import _UniverseStream
from refinitiv.data.delivery import omm_stream
from refinitiv.data.delivery._stream import StreamState
from tests.unit.conftest import StubSession
from tests.unit.delivery.stream.conftest import StubStreamConnection, event_mock

key = "ewff"
default_value = "EWGE#32"
value = "fgwFG3fd3F#Q"


def test_name_is_none():
    with pytest.raises(AttributeError):
        _UniverseStream(content_type=ContentType.NONE, name=None)


def test_with_argument():
    list_attr = [
        "_session",
        "_extended_params",
        "_stream",
    ]
    session = mock.Mock()
    session._omm_stream_counter = itertools.count(5)
    stream = _UniverseStream(
        content_type=ContentType.NONE, name="EUR=", session=session
    )
    for attr in list_attr:
        assert hasattr(stream, attr)
    assert stream._session is not None


def test_decode_partial_update_field():
    session = mock.Mock()
    sp = _UniverseStream(content_type=ContentType.NONE, name="", session=session)
    sp._record = {"Fields": dict()}
    assert sp._decode_partial_update_field(key=key, value=value) == value


def test_stream__repr__(stream):
    # given
    expected_value = "{'name': 'name', 'service': None, 'fields': {}}"

    # when
    testing_value = repr(stream)

    # then
    assert testing_value == expected_value


def test_stream__iter__(stream):
    # when
    iterator = iter(stream)

    # then
    with pytest.raises(StopIteration):
        next(iterator)


def test_stream__getitem__(stream):
    # then
    with pytest.raises(KeyError, match="not in Stream cache"):
        # when
        stream["item"]


def test_stream__len__(stream):
    # when
    testing_value = len(stream)

    # then
    assert testing_value == 0


def test_call_stream_close_for_unopened_stream(stream):
    # when
    state = stream.close()

    # then
    assert state is StreamState.Unopened, state


def test_stream_on_update(stream):
    # given
    callback_called = False

    def on_update(*owner):
        nonlocal callback_called
        callback_called = True

    stream.on_update(on_update)

    # when
    stream._stream._on_stream_update(dict(), {"Fields": {}})

    # then
    assert callback_called is True


def test_stream_on_status(stream):
    # given
    callback_called = False

    def on_status(owner, fields):
        nonlocal callback_called
        callback_called = True

    stream.on_status(on_status)

    # when
    stream._stream._on_stream_status(dict(), {"State": dict()})

    # then
    assert callback_called is True


def test_stream_on_complete(stream):
    # given
    callback_called = False

    def on_complete(owner):
        nonlocal callback_called
        callback_called = True

    stream.on_complete(on_complete)

    # when
    stream._stream._on_stream_complete(dict())

    # then
    assert callback_called is True


def test_stream_on_error(stream):
    # given
    callback_called = False

    def on_error(owner, fields):
        nonlocal callback_called
        callback_called = True

    stream.on_error(on_error)

    # when
    stream._stream._on_stream_error({}, {})

    # then
    assert callback_called is True


def test_stream_keys(stream):
    keys = stream.keys()

    assert keys is not None


def test_stream_values(stream):
    values = stream.values()

    assert values is not None


def test_stream_items(stream):
    items = stream.items()

    assert items is not None


def test_stream___iter___raise_error(stream):
    iterator = iter(stream)

    with pytest.raises(StopIteration):
        next(iterator)


def test_stream___iter___():
    session = StubSession(is_open=True)
    stream = _UniverseStream(
        content_type=ContentType.NONE,
        name="EUR=",
        fields=["field"],
        session=session,
    )

    iterator = iter(stream)
    testing_value = next(iterator)

    assert testing_value is not None


def test_stream___getitem__raise_error(stream):
    key = "key"

    with pytest.raises(KeyError, match=f"Field '{key}' not in Stream cache"):
        stream[key]


def test_stream___len__(stream):
    length = len(stream)

    assert length == 0


def test_stream___str__(stream):
    s = str(stream)

    assert s == "Unknown service|name[]"


@patch(
    "refinitiv.data.delivery._stream._stream_factory.create_stream_cxn",
    new=StubStreamConnection,
)
def test_ommstream_open_with_updates_true():
    # given
    session = StubSession(is_open=True)
    definition = omm_stream.Definition(name="name")
    stream = definition.get_stream(session)
    stream._stream._opened = event_mock

    # when
    stream.open(with_updates=True)

    # then
    assert stream._stream.open_message["Streaming"] is True


@patch(
    "refinitiv.data.delivery._stream._stream_factory.create_stream_cxn",
    new=StubStreamConnection,
)
def test_ommstream_open_with_updates_false():
    # given
    session = StubSession(is_open=True)
    definition = omm_stream.Definition(name="name")
    stream = definition.get_stream(session)
    stream._stream._opened = event_mock

    # when
    stream.open(with_updates=False)

    # then
    assert stream._stream.open_message["Streaming"] is False


@patch(
    "refinitiv.data.delivery._stream._stream_factory.create_stream_cxn",
    new=StubStreamConnection,
)
def test_ommstream_open_with_updates_from_true_to_false():
    # given
    session = StubSession(is_open=True)
    definition = omm_stream.Definition(name="name")
    stream = definition.get_stream(session)
    stream._stream._opened = event_mock

    # when
    stream.open(with_updates=True)
    stream.close()
    stream.open(with_updates=False)

    # then
    assert stream._stream.open_message["Streaming"] is False


@patch(
    "refinitiv.data.delivery._stream._stream_factory.create_stream_cxn",
    new=StubStreamConnection,
)
def test_ommstream_open_with_updates_from_false_to_true():
    # given
    session = StubSession(is_open=True)
    definition = omm_stream.Definition(name="name")
    stream = definition.get_stream(session)
    stream._stream._opened = event_mock

    # when
    stream.open(with_updates=False)
    stream.close()
    stream.open(with_updates=True)

    # then
    assert stream._stream.open_message["Streaming"] is True



import allure
import pytest

from refinitiv.data.content.ipa.surfaces import eti
from refinitiv.data.errors import RDError
from tests.integration.constants_list import HttpStatusCode, HttpReason
from tests.integration.content.ipa.conftest import (
    check_surface,
    check_surface_curve,
    check_surface_point,
)
from tests.integration.content.ipa.surfaces.eti.conftest import (
    surface_eti_definition,
    invalid_surface_eti_definition,
)
from tests.integration.helpers import (
    get_async_response_from_definitions,
    check_response_status,
    check_non_empty_response_data,
)


@allure.suite("Content object - Surface Eti")
@allure.feature("Content object - Surface Eti")
@allure.severity(allure.severity_level.CRITICAL)
class TestSurfaceEti:
    @allure.title(
        "Create a surface eti definition object with valid params and get data"
    )
    @pytest.mark.caseid("37821477")
    @pytest.mark.smoke
    def test_surface_eti_with_valid_params_and_get_data(self, open_session):
        response = surface_eti_definition().get_data()
        check_response_status(response, HttpStatusCode.TWO_HUNDRED, HttpReason.OK)
        check_non_empty_response_data(response)
        surface = response.data.surface
        check_surface(surface)
        curve = surface.get_curve("2022-03-18", eti.Axis.X)
        check_surface_curve(curve)
        point = surface.get_point("2022-03-18", 49.75)
        check_surface_point(point)

    @allure.title(
        "Create a surface eti definition object with valid params and get data asynchronously"
    )
    @pytest.mark.caseid("37821478")
    async def test_surface_eti_with_valid_params_and_get_data_async(self, open_session_async):
        valid_response, invalid_response = await get_async_response_from_definitions(
            surface_eti_definition(), invalid_surface_eti_definition()
        )
        surface = valid_response.data.surface
        check_surface(surface)
        curve = surface.get_curve("2021-12-17", eti.Axis.X)
        check_surface_curve(curve)
        point = surface.get_point("2021-12-17", 59.62)
        check_surface_point(point)

        check_response_status(
            response=invalid_response,
            expected_status_code=HttpStatusCode.TWO_HUNDRED,
            expected_http_reason=HttpReason.OK,
            expected_error_code="VolSurf.10006",
            expected_error_message="Invalid input: Instrument code must end with @RIC, @RICROOT",
        )

    @pytest.mark.xfail(reason="https://jira.refinitiv.com/browse/EAPI-4119")
    @allure.title("Create a surface eti - short definition")
    @pytest.mark.caseid("37821479")
    def test_surface_eti_with_simple_definition(self, open_session):
        response = eti.Definition(underlying_definition={"instrumentCode": "USD"}).get_data()
        surface = response.data.surface
        check_surface(surface)
        curve = surface.get_curve("2021-11-19", "X")
        check_surface_curve(curve)
        point = surface.get_point("2021-11-19", 50.89)
        check_surface_point(point)

    @allure.title("Create a invalid surface eti definition and get RDError")
    @pytest.mark.caseid("39606896")
    def test_invalid_surface_eti_definition(self, open_session):
        with pytest.raises(
            RDError,
            match="Error code VolSurf.10006 | Invalid input: Instrument code must end with @RIC, @RICROOT",
        ):
            invalid_surface_eti_definition().get_data()



import logging
import random

from refinitiv.data.delivery import cfs
from tests.integration.constants_list import ISO_DATE_FORMAT
from tests.integration.delivery.cfs.conftest import (
    convert_string_to_date_format,
)


def prepare_random_fileset_name(bucket_name):
    response = cfs.file_sets.Definition(bucket=bucket_name).get_data()
    random_fileset = random.choice(response.data.raw["value"])
    return random_fileset["name"]


def prepare_random_package_id(bucket_name):
    response = cfs.packages.Definition(bucket_name=bucket_name).get_data()
    random_package = random.choice(response.data.raw["value"])
    return random_package["packageId"]


def check_fileset_contains_all_attributes(fileset, expected_attributes_dict):
    def prepare_attribute_object(name, value):
        return {"name": name, "value": value}

    fileset_attributes = fileset["attributes"]
    for key in expected_attributes_dict:
        assert (
            prepare_attribute_object(key, expected_attributes_dict[key])
            in fileset_attributes
        ), f"Found fileset which does not contain expected attribute {key}: \n {fileset}"


def check_fileset_contains_package_id(fileset, expected_package_id):
    assert (
        fileset["packageId"].lower() == expected_package_id.lower()
    ), f"Found fileset which does not contain expected  package id {expected_package_id}: \n {fileset}"


def check_fileset_contains_status(fileset, expected_status):
    assert (
        fileset["status"].lower() == expected_status.lower()
    ), f"Found fileset which does not contain expected  package id {expected_status}: \n {fileset}"


def check_filesets_in_response_available_after_date(response, expected_date):
    date = convert_string_to_date_format(expected_date, ISO_DATE_FORMAT)
    filesets = response.data.raw["value"]
    for fileset in filesets:
        try:
            fileset_available_from_date = convert_string_to_date_format(
                fileset["availableFrom"], ISO_DATE_FORMAT
            )
            assert (
                fileset_available_from_date >= date
            ), f"found fileset available before the required date {date}: \n {fileset}"
        except KeyError:
            logging.warning(
                f"The fileset does not have 'availableFrom' field: {fileset}"
            )


def check_filesets_in_response_available_before_date(response, expected_date):
    date = convert_string_to_date_format(expected_date, ISO_DATE_FORMAT)
    filesets = response.data.raw["value"]
    for fileset in filesets:
        try:
            fileset_available_to_date = convert_string_to_date_format(
                fileset["availableTo"], ISO_DATE_FORMAT
            )
            assert (
                fileset_available_to_date <= date
            ), f"found fileset available after the required date {date}: \n {fileset}"
        except KeyError:
            logging.warning(f"The fileset does not have 'availableTo' field: {fileset}")


def check_filesets_in_response_have_content_to_param_before_date(
    response, expected_date
):
    date = convert_string_to_date_format(expected_date, ISO_DATE_FORMAT)
    filesets = response.data.raw["value"]
    for fileset in filesets:
        try:
            fileset_content_to_date = convert_string_to_date_format(
                fileset["contentTo"], ISO_DATE_FORMAT
            )
            assert (
                fileset_content_to_date <= date
            ), f"found fileset with contentTo property after the required date {date}: \n {fileset}"
        except KeyError:
            logging.warning(f"The fileset does not have 'contentTo' field: {fileset}")


def check_filesets_in_response_have_content_from_param_after_date(
    response, expected_date
):
    date = convert_string_to_date_format(expected_date, ISO_DATE_FORMAT)
    filesets = response.data.raw["value"]
    for fileset in filesets:
        try:
            fileset_content_from_date = convert_string_to_date_format(
                fileset["contentFrom"], ISO_DATE_FORMAT
            )
            assert (
                fileset_content_from_date >= date
            ), f"found fileset with contentFrom property before the required date {date}: \n {fileset}"
        except KeyError:
            logging.warning(f"The fileset does not have 'contentFrom' field: {fileset}")


def check_filesets_in_response_created_after_date(response, expected_date):
    date = convert_string_to_date_format(expected_date, ISO_DATE_FORMAT)
    filesets = response.data.raw["value"]
    for fileset in filesets:
        try:
            fileset_created_date = convert_string_to_date_format(
                fileset["created"], ISO_DATE_FORMAT
            )
            assert (
                fileset_created_date >= date
            ), f"found fileset created before the required date {date}: \n {fileset}"
        except KeyError:
            logging.warning(f"The fileset does not have 'created' field: {fileset}")


def check_filesets_in_response_modified_after_date(response, expected_date):
    date = convert_string_to_date_format(expected_date, ISO_DATE_FORMAT)
    filesets = response.data.raw["value"]
    for fileset in filesets:
        try:
            fileset_modified_date = convert_string_to_date_format(
                fileset["modified"], ISO_DATE_FORMAT
            )
            assert (
                fileset_modified_date >= date
            ), f"found fileset modified before the required date {date}: \n {fileset}"
        except KeyError:
            logging.warning(f"The fileset does not have 'modified' field: {fileset}")


def check_filesets_in_two_responses_are_not_the_same(response1, response2):
    for fileset in response1.data.raw["value"]:
        assert (
            fileset not in response2.data.raw["value"]
        ), f"Fileset from first request found in list of filesets in the second request: \n {fileset}"


def get_fileset_from_bucket_with_numfiles_bigger_than(bucket_name, num_files):
    response = cfs.file_sets.Definition(bucket_name, page_size=100).get_data()
    results = []
    for item in response.data.file_sets:
        if item["num_files"] > num_files:
            results.append(item)
    return random.choice(results)



import pytest
from refinitiv.data.content.ownership.consolidated import (
    breakdown,
    concentration,
    investors,
    recent_activity,
    shareholders_history_report,
    shareholders_report,
    top_n_concentration,
)


def test_breakdown():
    try:
        definition = breakdown.Definition(universe="", stat_type=1)
    except Exception as e:
        assert False, str(e)


def test_concentration():
    try:
        definition = concentration.Definition(universe="")
    except Exception as e:
        assert False, str(e)


def test_investors():
    try:
        definition = investors.Definition(universe="")
    except Exception as e:
        assert False, str(e)


def test_recent_activity():
    try:
        definition = recent_activity.Definition(universe="", sort_order="asc")
    except Exception as e:
        assert False, str(e)


def test_shareholders_history_report():
    try:
        definition = shareholders_history_report.Definition(universe="", frequency="Q")
    except Exception as e:
        assert False, str(e)


def test_shareholders_history_report_universe_is_list():
    with pytest.raises(ValueError):
        definition = shareholders_history_report.Definition(universe=[], frequency="Q")


def test_shareholders_report():
    try:
        definition = shareholders_report.Definition(universe="")
    except Exception as e:
        assert False, str(e)


def test_top_n_concentration():
    try:
        definition = top_n_concentration.Definition(universe="", count=10)
    except Exception as e:
        assert False, str(e)



# coding: utf8

from typing import Optional, List

from ....._tools import try_copy_to_list
from ..._object_definition import ObjectDefinition

from ._enums import (
    FxSwapCalculationMethod,
    OptionVolatilityType,
    PriceSide,
    PricingModelType,
    TimeStamp,
    VolatilityModel,
)
from ._models import (
    BidAskMid,
    InterpolationWeight,
    PayoutScaling,
)


class PricingParameters(ObjectDefinition):
    """
    API endpoint for Financial Contract analytics,
    that returns calculations relevant to each contract type.

    Parameters
    ----------
    atm_volatility_object : BidAskMid, optional

    butterfly10_d_object : BidAskMid, optional

    butterfly25_d_object : BidAskMid, optional

    domestic_deposit_rate_percent_object : BidAskMid, optional

    foreign_deposit_rate_percent_object : BidAskMid, optional

    forward_points_object : BidAskMid, optional

    fx_spot_object : BidAskMid, optional

    fx_swap_calculation_method : FxSwapCalculationMethod, optional
        The method used to calculate an outright price or deposit rates. the possible
        values are:
        - fxswapimpliedfromdeposit: implied fx swap points are computed from deposit
          rates.
        - depositccy1impliedfromfxswap: currency 1 deposit rates are computed using swap
          points,
        - depositccy2impliedfromfxswap: currency 2 deposit rates are computed using swap
          points.  the default value is 'depositccy2impliedfromfxswap'.
    implied_volatility_object : BidAskMid, optional

    interpolation_weight : InterpolationWeight, optional

    option_price_side : PriceSide, optional
        The quoted price side of the instrument.the possible values are:
        - bid,
        - ask,
        - mid,
        - last. optional. the default values for listed options are:
        - ask: if buysell is set to 'buy',
        - bid: if buysell is set to 'sell',
        - last: if buysell is not provided. the default value for otc options is 'mid'.
    option_time_stamp : TimeStamp, optional
        The mode of the instrument's timestamp selection. the possible values are:
        - open: the opening value of valuationdate, or if it is not available, the close
          of the previous day is used,
        - close: the close value of valuationdate is used,
        - default: the latest snapshot is used when valuationdate is today, and the
          close price when valuationdate is in the past. optional.the default value is
          'default'.
    payout_custom_dates : string, optional
        The array of dates set by a user for the payout/volatility chart. optional.no
        default value applies.
    payout_scaling_interval : PayoutScaling, optional

    price_side : PriceSide, optional
        The quoted price side of the instrument. the possible values are:
        - bid,
        - ask,
        - mid.
    pricing_model_type : PricingModelType, optional
        The model type of the option pricing. the possible values are:
        - blackscholes
        - bachelier (available for commodity options including calendar spread options)
        - whaley
        - binomial
        - trinomial
        - localvolatility (applicable only for barrier options, cbbc options and binary
          options)
        - vannavolga (only applicable for fxbarrieroption, fxdigitaloption and
          fxtouchesoption)  optional. the default value depends on the option type.
    risk_reversal10_d_object : BidAskMid, optional

    risk_reversal25_d_object : BidAskMid, optional

    underlying_price_side : PriceSide, optional
        The quoted price side of the underlying asset.the possible values are:
        - bid,
        - ask,
        - mid,
        - last. optional. the default values are:
        - ask: if buysell is set to 'buy',
        - bid: if buysell is set to 'sell',
        - last: if buysell is not provided.
    underlying_time_stamp : TimeStamp, optional
        The mode of the underlying asset's timestamp selection. the possible values are:
        - open: the opening value of valuationdate, or if it is not available, the close
          of the previous day is used,
        - close: the close value of valuationdate is used,
        - default: the latest snapshot is used when valuationdate is today, and the
          close price when valuationdate is in the past. optional.the default value is
          'default'.
    volatility_model : VolatilityModel, optional
        The model used to build the volatility surface. the possible values are:
        - sabr,
        - cubicspline,
        - svi,
        - twinlognormal,
        - vannavolga10d,
        - vannavolga25d.
    volatility_type : OptionVolatilityType, optional
        The type of volatility for the option pricing. the possible values are:
        - implied: the volatility anticipated for the underlying asset for the remaining
          life of the option(implied by an option premium),
        - svisurface: the volatility computed from volsurface service,
        - historical: the volatility of the underlying asset during a period in the
          past. the value 'implied' is available only for listed options. if
          volatilitypercent is defined, volatilitytype is not taken into account.
          optional. the default value is 'implied'.
    compute_payout_chart : bool, optional
        Define whether the payout chart must be computed or not
    compute_volatility_payout : bool, optional
        Define whether the volatility payout chart must be computed or not
    cutoff_time : str, optional
        The cutoff time
    cutoff_time_zone : str, optional
        The cutoff time zone
    market_data_date : str, optional
        The date at which the market data is retrieved. the value is expressed in iso
        8601 format: yyyy-mm-ddt[hh]:[mm]:[ss]z (e.g., '2021-01-01t00:00:00z'). it
        should be less or equal tovaluationdate). optional. by
        default,marketdatadateisvaluationdateor today.
    market_value_in_deal_ccy : float, optional
        The market value (premium) of the instrument. the value is expressed in the deal
        currency. it is used to define optionprice and compute volatilitypercent. if
        marketvalueindealccy is defined, optionpriceside and volatilitypercent are not
        taken into account; marketvalueindealccy and marketvalueinreportccy cannot be
        overriden at a time. optional. by default, it is equal to optionprice for listed
        options or computed from volatilitypercent for otc options.
    market_value_in_report_ccy : float, optional
        The market value (premium) of the instrument. it is computed as
        [marketvalueindealccy  fxspot]. the value is expressed in the reporting
        currency. it is used to define optionprice and computevolatilitypercent.
        ifmarketvalueinreportccyis defined, optionpriceside and volatilitypercentinputs
        are not taken into account; marketvalueindealccy and marketvalueinreportccy
        cannot be overriden at a time. optional. by default, fxspot rate is retrieved
        from the market data.
    report_ccy : str, optional
        The reporting currency code, expressed in iso 4217 alphabetical format (e.g.,
        'usd'). it is set for the fields ending with 'xxxinreportccy'. optional. the
        default value is the notional currency.
    report_ccy_rate : float, optional
        The rate of the reporting currency against the option currency. it can be used
        to calculate optionprice and marketvalueindealccy if marketvalueinreportccy is
        defined. optional.by default, it is retrieved from the market data.
    risk_free_rate_percent : float, optional
        A risk-free rate of the option currency used for the option pricing. optional.
        by default, the value is retrieved from the market data.
    simulate_exercise : bool, optional
        Tells if payoff-linked cashflow should be returned. possible values:
        - true
        - false
    underlying_price : float, optional
        The price of the underlying asset. the value is expressed in the deal currency.
        if underlyingprice is defined, underlyingpriceside is not taken into account.
        optional. by default, the value is retrieved from the market data.
    valuation_date : str, optional
        The date at which the instrument is valued. the value is expressed in iso 8601
        format: yyyy-mm-ddt[hh]:[mm]:[ss]z (e.g., '2021-01-01t00:00:00z'). by default,
        marketdatadate is used. if marketdatadate is not specified, the default value is
        today.
    volatility : float, optional
        Volatility(without unity) to override and that will be used as pricing analysis
        input to compute marketvalueindealccy. introduced due to bachelier model, for
        more details please have a look at apqps-13558 optional. no override is applied
        by default. note that if premium is defined, volatility is not taken into
        account.
    volatility_percent : float, optional
        The degree of the underlying asset's price variations over a specified time
        period, used for the option pricing. the value is expressed in percentages. it
        is used to compute marketvalueindealccy.if marketvalueindealccy is defined,
        volatilitypercent is not taken into account. optional. by default, it is
        computed from marketvalueindealccy. if volsurface fails to return a volatility,
        it defaults to '20'.
    """

    def __init__(
        self,
        atm_volatility_object: Optional[BidAskMid] = None,
        butterfly10_d_object: Optional[BidAskMid] = None,
        butterfly25_d_object: Optional[BidAskMid] = None,
        domestic_deposit_rate_percent_object: Optional[BidAskMid] = None,
        foreign_deposit_rate_percent_object: Optional[BidAskMid] = None,
        forward_points_object: Optional[BidAskMid] = None,
        fx_spot_object: Optional[BidAskMid] = None,
        fx_swap_calculation_method: Optional[FxSwapCalculationMethod] = None,
        implied_volatility_object: Optional[BidAskMid] = None,
        interpolation_weight: Optional[InterpolationWeight] = None,
        option_price_side: Optional[PriceSide] = None,
        option_time_stamp: Optional[TimeStamp] = None,
        payout_custom_dates: Optional[List[str]] = None,
        payout_scaling_interval: Optional[PayoutScaling] = None,
        price_side: Optional[PriceSide] = None,
        pricing_model_type: Optional[PricingModelType] = None,
        risk_reversal10_d_object: Optional[BidAskMid] = None,
        risk_reversal25_d_object: Optional[BidAskMid] = None,
        underlying_price_side: Optional[PriceSide] = None,
        underlying_time_stamp: Optional[TimeStamp] = None,
        volatility_model: Optional[VolatilityModel] = None,
        volatility_type: Optional[OptionVolatilityType] = None,
        compute_payout_chart: Optional[bool] = None,
        compute_volatility_payout: Optional[bool] = None,
        cutoff_time: Optional[str] = None,
        cutoff_time_zone: Optional[str] = None,
        market_data_date: Optional[str] = None,
        market_value_in_deal_ccy: Optional[float] = None,
        market_value_in_report_ccy: Optional[float] = None,
        report_ccy: Optional[str] = None,
        report_ccy_rate: Optional[float] = None,
        risk_free_rate_percent: Optional[float] = None,
        simulate_exercise: Optional[bool] = None,
        underlying_price: Optional[float] = None,
        valuation_date: Optional[str] = None,
        volatility: Optional[float] = None,
        volatility_percent: Optional[float] = None,
    ) -> None:
        super().__init__()
        self.atm_volatility_object = atm_volatility_object
        self.butterfly10_d_object = butterfly10_d_object
        self.butterfly25_d_object = butterfly25_d_object
        self.domestic_deposit_rate_percent_object = domestic_deposit_rate_percent_object
        self.foreign_deposit_rate_percent_object = foreign_deposit_rate_percent_object
        self.forward_points_object = forward_points_object
        self.fx_spot_object = fx_spot_object
        self.fx_swap_calculation_method = fx_swap_calculation_method
        self.implied_volatility_object = implied_volatility_object
        self.interpolation_weight = interpolation_weight
        self.option_price_side = option_price_side
        self.option_time_stamp = option_time_stamp
        self.payout_custom_dates = try_copy_to_list(payout_custom_dates)
        self.payout_scaling_interval = payout_scaling_interval
        self.price_side = price_side
        self.pricing_model_type = pricing_model_type
        self.risk_reversal10_d_object = risk_reversal10_d_object
        self.risk_reversal25_d_object = risk_reversal25_d_object
        self.underlying_price_side = underlying_price_side
        self.underlying_time_stamp = underlying_time_stamp
        self.volatility_model = volatility_model
        self.volatility_type = volatility_type
        self.compute_payout_chart = compute_payout_chart
        self.compute_volatility_payout = compute_volatility_payout
        self.cutoff_time = cutoff_time
        self.cutoff_time_zone = cutoff_time_zone
        self.market_data_date = market_data_date
        self.market_value_in_deal_ccy = market_value_in_deal_ccy
        self.market_value_in_report_ccy = market_value_in_report_ccy
        self.report_ccy = report_ccy
        self.report_ccy_rate = report_ccy_rate
        self.risk_free_rate_percent = risk_free_rate_percent
        self.simulate_exercise = simulate_exercise
        self.underlying_price = underlying_price
        self.valuation_date = valuation_date
        self.volatility = volatility
        self.volatility_percent = volatility_percent

    @property
    def atm_volatility_object(self):
        """
        :return: object BidAskMid
        """
        return self._get_object_parameter(BidAskMid, "atmVolatilityObject")

    @atm_volatility_object.setter
    def atm_volatility_object(self, value):
        self._set_object_parameter(BidAskMid, "atmVolatilityObject", value)

    @property
    def butterfly10_d_object(self):
        """
        :return: object BidAskMid
        """
        return self._get_object_parameter(BidAskMid, "butterfly10DObject")

    @butterfly10_d_object.setter
    def butterfly10_d_object(self, value):
        self._set_object_parameter(BidAskMid, "butterfly10DObject", value)

    @property
    def butterfly25_d_object(self):
        """
        :return: object BidAskMid
        """
        return self._get_object_parameter(BidAskMid, "butterfly25DObject")

    @butterfly25_d_object.setter
    def butterfly25_d_object(self, value):
        self._set_object_parameter(BidAskMid, "butterfly25DObject", value)

    @property
    def domestic_deposit_rate_percent_object(self):
        """
        :return: object BidAskMid
        """
        return self._get_object_parameter(BidAskMid, "domesticDepositRatePercentObject")

    @domestic_deposit_rate_percent_object.setter
    def domestic_deposit_rate_percent_object(self, value):
        self._set_object_parameter(BidAskMid, "domesticDepositRatePercentObject", value)

    @property
    def foreign_deposit_rate_percent_object(self):
        """
        :return: object BidAskMid
        """
        return self._get_object_parameter(BidAskMid, "foreignDepositRatePercentObject")

    @foreign_deposit_rate_percent_object.setter
    def foreign_deposit_rate_percent_object(self, value):
        self._set_object_parameter(BidAskMid, "foreignDepositRatePercentObject", value)

    @property
    def forward_points_object(self):
        """
        :return: object BidAskMid
        """
        return self._get_object_parameter(BidAskMid, "forwardPointsObject")

    @forward_points_object.setter
    def forward_points_object(self, value):
        self._set_object_parameter(BidAskMid, "forwardPointsObject", value)

    @property
    def fx_spot_object(self):
        """
        :return: object BidAskMid
        """
        return self._get_object_parameter(BidAskMid, "fxSpotObject")

    @fx_spot_object.setter
    def fx_spot_object(self, value):
        self._set_object_parameter(BidAskMid, "fxSpotObject", value)

    @property
    def fx_swap_calculation_method(self):
        """
        The method used to calculate an outright price or deposit rates. the possible
        values are:
        - fxswapimpliedfromdeposit: implied fx swap points are computed from deposit
          rates.
        - depositccy1impliedfromfxswap: currency 1 deposit rates are computed using swap
          points,
        - depositccy2impliedfromfxswap: currency 2 deposit rates are computed using swap
          points.  the default value is 'depositccy2impliedfromfxswap'.
        :return: enum FxSwapCalculationMethod
        """
        return self._get_enum_parameter(
            FxSwapCalculationMethod, "fxSwapCalculationMethod"
        )

    @fx_swap_calculation_method.setter
    def fx_swap_calculation_method(self, value):
        self._set_enum_parameter(
            FxSwapCalculationMethod, "fxSwapCalculationMethod", value
        )

    @property
    def implied_volatility_object(self):
        """
        :return: object BidAskMid
        """
        return self._get_object_parameter(BidAskMid, "impliedVolatilityObject")

    @implied_volatility_object.setter
    def implied_volatility_object(self, value):
        self._set_object_parameter(BidAskMid, "impliedVolatilityObject", value)

    @property
    def interpolation_weight(self):
        """
        :return: object InterpolationWeight
        """
        return self._get_object_parameter(InterpolationWeight, "interpolationWeight")

    @interpolation_weight.setter
    def interpolation_weight(self, value):
        self._set_object_parameter(InterpolationWeight, "interpolationWeight", value)

    @property
    def option_price_side(self):
        """
        The quoted price side of the instrument.the possible values are:
        - bid,
        - ask,
        - mid,
        - last. optional. the default values for listed options are:
        - ask: if buysell is set to 'buy',
        - bid: if buysell is set to 'sell',
        - last: if buysell is not provided. the default value for otc options is 'mid'.
        :return: enum PriceSide
        """
        return self._get_enum_parameter(PriceSide, "optionPriceSide")

    @option_price_side.setter
    def option_price_side(self, value):
        self._set_enum_parameter(PriceSide, "optionPriceSide", value)

    @property
    def option_time_stamp(self):
        """
        The mode of the instrument's timestamp selection. the possible values are:
        - open: the opening value of valuationdate, or if it is not available, the close
          of the previous day is used,
        - close: the close value of valuationdate is used,
        - default: the latest snapshot is used when valuationdate is today, and the
          close price when valuationdate is in the past. optional.the default value is
          'default'.
        :return: enum TimeStamp
        """
        return self._get_enum_parameter(TimeStamp, "optionTimeStamp")

    @option_time_stamp.setter
    def option_time_stamp(self, value):
        self._set_enum_parameter(TimeStamp, "optionTimeStamp", value)

    @property
    def payout_custom_dates(self):
        """
        The array of dates set by a user for the payout/volatility chart. optional.no
        default value applies.
        :return: list string
        """
        return self._get_list_parameter(str, "payoutCustomDates")

    @payout_custom_dates.setter
    def payout_custom_dates(self, value):
        self._set_list_parameter(str, "payoutCustomDates", value)

    @property
    def payout_scaling_interval(self):
        """
        :return: object PayoutScaling
        """
        return self._get_object_parameter(PayoutScaling, "payoutScalingInterval")

    @payout_scaling_interval.setter
    def payout_scaling_interval(self, value):
        self._set_object_parameter(PayoutScaling, "payoutScalingInterval", value)

    @property
    def price_side(self):
        """
        The quoted price side of the instrument. the possible values are:
        - bid,
        - ask,
        - mid.
        :return: enum PriceSide
        """
        return self._get_enum_parameter(PriceSide, "priceSide")

    @price_side.setter
    def price_side(self, value):
        self._set_enum_parameter(PriceSide, "priceSide", value)

    @property
    def pricing_model_type(self):
        """
        The model type of the option pricing. the possible values are:
        - blackscholes
        - bachelier (available for commodity options including calendar spread options)
        - whaley
        - binomial
        - trinomial
        - localvolatility (applicable only for barrier options, cbbc options and binary
          options)
        - vannavolga (only applicable for fxbarrieroption, fxdigitaloption and
          fxtouchesoption)  optional. the default value depends on the option type.
        :return: enum PricingModelType
        """
        return self._get_enum_parameter(PricingModelType, "pricingModelType")

    @pricing_model_type.setter
    def pricing_model_type(self, value):
        self._set_enum_parameter(PricingModelType, "pricingModelType", value)

    @property
    def risk_reversal10_d_object(self):
        """
        :return: object BidAskMid
        """
        return self._get_object_parameter(BidAskMid, "riskReversal10DObject")

    @risk_reversal10_d_object.setter
    def risk_reversal10_d_object(self, value):
        self._set_object_parameter(BidAskMid, "riskReversal10DObject", value)

    @property
    def risk_reversal25_d_object(self):
        """
        :return: object BidAskMid
        """
        return self._get_object_parameter(BidAskMid, "riskReversal25DObject")

    @risk_reversal25_d_object.setter
    def risk_reversal25_d_object(self, value):
        self._set_object_parameter(BidAskMid, "riskReversal25DObject", value)

    @property
    def underlying_price_side(self):
        """
        The quoted price side of the underlying asset.the possible values are:
        - bid,
        - ask,
        - mid,
        - last. optional. the default values are:
        - ask: if buysell is set to 'buy',
        - bid: if buysell is set to 'sell',
        - last: if buysell is not provided.
        :return: enum PriceSide
        """
        return self._get_enum_parameter(PriceSide, "underlyingPriceSide")

    @underlying_price_side.setter
    def underlying_price_side(self, value):
        self._set_enum_parameter(PriceSide, "underlyingPriceSide", value)

    @property
    def underlying_time_stamp(self):
        """
        The mode of the underlying asset's timestamp selection. the possible values are:
        - open: the opening value of valuationdate, or if it is not available, the close
          of the previous day is used,
        - close: the close value of valuationdate is used,
        - default: the latest snapshot is used when valuationdate is today, and the
          close price when valuationdate is in the past. optional.the default value is
          'default'.
        :return: enum TimeStamp
        """
        return self._get_enum_parameter(TimeStamp, "underlyingTimeStamp")

    @underlying_time_stamp.setter
    def underlying_time_stamp(self, value):
        self._set_enum_parameter(TimeStamp, "underlyingTimeStamp", value)

    @property
    def volatility_model(self):
        """
        The model used to build the volatility surface. the possible values are:
        - sabr,
        - cubicspline,
        - svi,
        - twinlognormal,
        - vannavolga10d,
        - vannavolga25d.
        :return: enum VolatilityModel
        """
        return self._get_enum_parameter(VolatilityModel, "volatilityModel")

    @volatility_model.setter
    def volatility_model(self, value):
        self._set_enum_parameter(VolatilityModel, "volatilityModel", value)

    @property
    def volatility_type(self):
        """
        The type of volatility for the option pricing. the possible values are:
        - implied: the volatility anticipated for the underlying asset for the remaining
          life of the option(implied by an option premium),
        - svisurface: the volatility computed from volsurface service,
        - historical: the volatility of the underlying asset during a period in the
          past. the value 'implied' is available only for listed options. if
          volatilitypercent is defined, volatilitytype is not taken into account.
          optional. the default value is 'implied'.
        :return: enum OptionVolatilityType
        """
        return self._get_enum_parameter(OptionVolatilityType, "volatilityType")

    @volatility_type.setter
    def volatility_type(self, value):
        self._set_enum_parameter(OptionVolatilityType, "volatilityType", value)

    @property
    def compute_payout_chart(self):
        """
        Define whether the payout chart must be computed or not
        :return: bool
        """
        return self._get_parameter("computePayoutChart")

    @compute_payout_chart.setter
    def compute_payout_chart(self, value):
        self._set_parameter("computePayoutChart", value)

    @property
    def compute_volatility_payout(self):
        """
        Define whether the volatility payout chart must be computed or not
        :return: bool
        """
        return self._get_parameter("computeVolatilityPayout")

    @compute_volatility_payout.setter
    def compute_volatility_payout(self, value):
        self._set_parameter("computeVolatilityPayout", value)

    @property
    def cutoff_time(self):
        """
        The cutoff time
        :return: str
        """
        return self._get_parameter("cutoffTime")

    @cutoff_time.setter
    def cutoff_time(self, value):
        self._set_parameter("cutoffTime", value)

    @property
    def cutoff_time_zone(self):
        """
        The cutoff time zone
        :return: str
        """
        return self._get_parameter("cutoffTimeZone")

    @cutoff_time_zone.setter
    def cutoff_time_zone(self, value):
        self._set_parameter("cutoffTimeZone", value)

    @property
    def market_data_date(self):
        """
        The date at which the market data is retrieved. the value is expressed in iso
        8601 format: yyyy-mm-ddt[hh]:[mm]:[ss]z (e.g., '2021-01-01t00:00:00z'). it
        should be less or equal tovaluationdate). optional. by
        default,marketdatadateisvaluationdateor today.
        :return: str
        """
        return self._get_parameter("marketDataDate")

    @market_data_date.setter
    def market_data_date(self, value):
        self._set_parameter("marketDataDate", value)

    @property
    def market_value_in_deal_ccy(self):
        """
        The market value (premium) of the instrument. the value is expressed in the deal
        currency. it is used to define optionprice and compute volatilitypercent. if
        marketvalueindealccy is defined, optionpriceside and volatilitypercent are not
        taken into account; marketvalueindealccy and marketvalueinreportccy cannot be
        overriden at a time. optional. by default, it is equal to optionprice for listed
        options or computed from volatilitypercent for otc options.
        :return: float
        """
        return self._get_parameter("marketValueInDealCcy")

    @market_value_in_deal_ccy.setter
    def market_value_in_deal_ccy(self, value):
        self._set_parameter("marketValueInDealCcy", value)

    @property
    def market_value_in_report_ccy(self):
        """
        The market value (premium) of the instrument. it is computed as
        [marketvalueindealccy  fxspot]. the value is expressed in the reporting
        currency. it is used to define optionprice and computevolatilitypercent.
        ifmarketvalueinreportccyis defined, optionpriceside and volatilitypercentinputs
        are not taken into account; marketvalueindealccy and marketvalueinreportccy
        cannot be overriden at a time. optional. by default, fxspot rate is retrieved
        from the market data.
        :return: float
        """
        return self._get_parameter("marketValueInReportCcy")

    @market_value_in_report_ccy.setter
    def market_value_in_report_ccy(self, value):
        self._set_parameter("marketValueInReportCcy", value)

    @property
    def report_ccy(self):
        """
        The reporting currency code, expressed in iso 4217 alphabetical format (e.g.,
        'usd'). it is set for the fields ending with 'xxxinreportccy'. optional. the
        default value is the notional currency.
        :return: str
        """
        return self._get_parameter("reportCcy")

    @report_ccy.setter
    def report_ccy(self, value):
        self._set_parameter("reportCcy", value)

    @property
    def report_ccy_rate(self):
        """
        The rate of the reporting currency against the option currency. it can be used
        to calculate optionprice and marketvalueindealccy if marketvalueinreportccy is
        defined. optional.by default, it is retrieved from the market data.
        :return: float
        """
        return self._get_parameter("reportCcyRate")

    @report_ccy_rate.setter
    def report_ccy_rate(self, value):
        self._set_parameter("reportCcyRate", value)

    @property
    def risk_free_rate_percent(self):
        """
        A risk-free rate of the option currency used for the option pricing. optional.
        by default, the value is retrieved from the market data.
        :return: float
        """
        return self._get_parameter("riskFreeRatePercent")

    @risk_free_rate_percent.setter
    def risk_free_rate_percent(self, value):
        self._set_parameter("riskFreeRatePercent", value)

    @property
    def simulate_exercise(self):
        """
        Tells if payoff-linked cashflow should be returned. possible values:
        - true
        - false
        :return: bool
        """
        return self._get_parameter("simulateExercise")

    @simulate_exercise.setter
    def simulate_exercise(self, value):
        self._set_parameter("simulateExercise", value)

    @property
    def underlying_price(self):
        """
        The price of the underlying asset. the value is expressed in the deal currency.
        if underlyingprice is defined, underlyingpriceside is not taken into account.
        optional. by default, the value is retrieved from the market data.
        :return: float
        """
        return self._get_parameter("underlyingPrice")

    @underlying_price.setter
    def underlying_price(self, value):
        self._set_parameter("underlyingPrice", value)

    @property
    def valuation_date(self):
        """
        The date at which the instrument is valued. the value is expressed in iso 8601
        format: yyyy-mm-ddt[hh]:[mm]:[ss]z (e.g., '2021-01-01t00:00:00z'). by default,
        marketdatadate is used. if marketdatadate is not specified, the default value is
        today.
        :return: str
        """
        return self._get_parameter("valuationDate")

    @valuation_date.setter
    def valuation_date(self, value):
        self._set_parameter("valuationDate", value)

    @property
    def volatility(self):
        """
        Volatility(without unity) to override and that will be used as pricing analysis
        input to compute marketvalueindealccy. introduced due to bachelier model, for
        more details please have a look at apqps-13558 optional. no override is applied
        by default. note that if premium is defined, volatility is not taken into
        account.
        :return: float
        """
        return self._get_parameter("volatility")

    @volatility.setter
    def volatility(self, value):
        self._set_parameter("volatility", value)

    @property
    def volatility_percent(self):
        """
        The degree of the underlying asset's price variations over a specified time
        period, used for the option pricing. the value is expressed in percentages. it
        is used to compute marketvalueindealccy.if marketvalueindealccy is defined,
        volatilitypercent is not taken into account. optional. by default, it is
        computed from marketvalueindealccy. if volsurface fails to return a volatility,
        it defaults to '20'.
        :return: float
        """
        return self._get_parameter("volatilityPercent")

    @volatility_percent.setter
    def volatility_percent(self, value):
        self._set_parameter("volatilityPercent", value)



from typing import TYPE_CHECKING, Callable, Any

from ._protocol_type import ProtocolType
from ._stream_listener import RDPStreamListener
from .stream import Stream, update_message_with_extended_params
from ..._tools import cached_property

if TYPE_CHECKING:
    from ._stream_factory import StreamDetails
    from . import StreamConnection
    from ..._types import ExtendedParams
    from ..._core.session import Session


class _RDPStream(Stream, RDPStreamListener["_RDPStream"]):
    def __init__(
        self,
        stream_id: int,
        session: "Session",
        service: str,
        universe: list,
        view: list,
        parameters: dict,
        details: "StreamDetails",
        extended_params: "ExtendedParams" = None,
        on_ack: Callable = None,
        on_response: Callable = None,
        on_update: Callable = None,
        on_alarm: Callable = None,
    ):
        Stream.__init__(self, stream_id, session, details)
        RDPStreamListener.__init__(
            self,
            logger=session.logger(),
            on_ack=on_ack,
            on_response=on_response,
            on_update=on_update,
            on_alarm=on_alarm,
        )

        self._service = service
        self._universe = universe
        self._view = view
        self._parameters = parameters
        self._extended_params = extended_params

    @property
    def service(self) -> str:
        return self._service

    @property
    def universe(self) -> list:
        return self._universe

    @property
    def view(self) -> list:
        return self._view

    @property
    def parameters(self) -> dict:
        return self._parameters

    @property
    def extended_params(self) -> "ExtendedParams":
        return self._extended_params

    @property
    def name(self) -> str:
        return str(self._universe)

    @property
    def protocol_type(self) -> ProtocolType:
        return ProtocolType.RDP

    @cached_property
    def close_message(self):
        return {"streamID": f"{self.id:d}", "method": "Close"}

    @cached_property
    def open_message(self) -> dict:
        message = {
            "streamID": f"{self.id:d}",
            "method": "Subscribe",
            "universe": self.universe,
        }

        if self.service is not None:
            message["service"] = self.service

        if self.view is not None:
            message["view"] = self.view

        if self.parameters is not None:
            message["parameters"] = self.parameters

        if self.extended_params:
            message = update_message_with_extended_params(message, self.extended_params)

        return message

    def _do_open(self, *args, **kwargs) -> None:
        self._initialize_cxn()

        self._cxn.on(self._event.ack_by_id, self._on_stream_ack)
        self._cxn.on(self._event.update_by_id, self._on_stream_update)
        self._cxn.on(self._event.response_by_id, self._on_stream_response)
        self._cxn.on(self._event.alarm_by_id, self._on_stream_alarm)

        super()._do_open(*args, **kwargs)

    def _dispose(self):
        self._debug(f"{self._classname} disposing [d]")
        if self._cxn is not None:
            self._cxn.remove_listener(self._event.ack_by_id, self._on_stream_ack)
            self._cxn.remove_listener(self._event.update_by_id, self._on_stream_update)
            self._cxn.remove_listener(
                self._event.response_by_id, self._on_stream_response
            )
            self._cxn.remove_listener(self._event.alarm_by_id, self._on_stream_alarm)

            self._release_cxn()

        self._debug(f"{self._classname} disposed [D]")

    def _do_on_stream_response(self, originator: "StreamConnection", *args) -> Any:
        if self.is_opening:
            self._opened.set()

        return args

    def _do_on_stream_ack(self, originator: "StreamConnection", *args) -> Any:
        message = args[0]
        message_state = message.get("state", {})
        stream_state = message_state.get("stream")

        if stream_state == "Closed":
            self._debug(
                f"{self._classname} received a closing message, "
                f"message_state={message_state}, state={self.state}"
            )

            if self.is_opening:
                self._opened.set()

        return args

    def _do_on_stream_alarm(self, originator, *args) -> Any:
        """
        {
            "data": [],
            "state": {
                "id": "QPSValuation.ERROR_REQUEST_TIMEOUT",
                "code": 408,
                "status": "ERROR",
                "message": "The request could not be executed
                            within the service allocated time",
                "stream": "Open"
            },
            "type": "Alarm",
            "streamID": "3"
        }
        """
        message = args[0]
        self._error(f"{message}")

        if self.is_opening:
            self._opened.set()

        return args



from ..._core.session import get_default
from ..._errors import RDError
from ..._fin_coder_layer import get_adc_data
from ..._tools import cached_property
from ._chain import default_error_message
from ._universe_expander import UniverseExpander


def update_universe(raw, _universe):
    index = 0  # instrument
    data = raw.get("data")
    if data and all(isinstance(i[index], str) for i in data):
        universe = [i[index] for i in data]
    else:
        universe = _universe
    return universe


def get_universe(expression):
    session = get_default()
    logger = session.logger()
    adc_raw, _ = get_adc_data(
        params={
            "universe": expression,
            "fields": "TR.RIC",
        },
        logger=logger,
        raise_if_error=True,
    )
    return update_universe(
        adc_raw,
        None,
    )


class DiscoveryUniverse(UniverseExpander):
    def __init__(self, expression):
        self._expression = expression

    @property
    def expression(self):
        return self._expression

    @cached_property
    def _universe(self):
        universe = get_universe(self._expression)
        if not universe:
            raise RDError(-1, default_error_message)
        return universe



import pytest

from refinitiv.data.content.news._tools import (
    news_build_df_rdp,
    _get_text_from_story,
    _get_headline_from_story,
)


@pytest.mark.parametrize(
    ("input_data",),
    [
        (
            {
                "data": [
                    {
                        "storyId": "urn:newsml:reuters.com:20211229:nNDL1r6YQr:1",
                        "newsItem": {
                            "contentMeta": {
                                "infoSource": [
                                    {"_qcode": "NS:PUBT", "_role": "sRole:source"},
                                    {"_qcode": "NS:PUBT", "_role": "sRole:origProv"},
                                ],
                            },
                            "itemMeta": {
                                "firstCreated": {"$": "2021-12-29T12:15:06Z"},
                                "versionCreated": {"$": "2021-12-29T12:15:06Z"},
                                "title": [
                                    {
                                        "$": "HubSpot Inc. - 26 Companies With Really Catchy Slogans & Brand Taglines"
                                    }
                                ],
                            },
                        },
                    },
                    {
                        "storyId": "urn:newsml:reuters.com:20211229:nDjc3FXTxC:1",
                        "newsItem": {
                            "_version": 1,
                            "contentMeta": {
                                "infoSource": [
                                    {"_qcode": "NS:DJCP", "_role": "sRole:source"},
                                    {"_qcode": "NS:DJCP", "_role": "sRole:origProv"},
                                ],
                            },
                            "itemMeta": {
                                "firstCreated": {"$": "2021-12-29T00:00:00Z"},
                                "versionCreated": {"$": "2021-12-29T11:30:00Z"},
                                "title": [
                                    {
                                        "$": "The Public Safety Analytics Market is Expected to Witness a CAGR of Over 29.39% During the Forecast Period (2021 - 2026)"
                                    }
                                ],
                            },
                        },
                    },
                    {
                        "storyId": "urn:newsml:reuters.com:20211229:nDjc7scRMV:1",
                        "newsItem": {
                            "_version": 1,
                            "contentMeta": {
                                "infoSource": [
                                    {"_qcode": "NS:DJCP", "_role": "sRole:source"},
                                    {"_qcode": "NS:DJCP", "_role": "sRole:origProv"},
                                ],
                            },
                            "itemMeta": {
                                "firstCreated": {"$": "2021-12-29T00:00:00Z"},
                                "versionCreated": {"$": "2021-12-29T11:15:00Z"},
                                "title": [
                                    {
                                        "$": "Global Healthcare CRM Market 2021-2026: Key Players Include Accenture, Amdocs, Microsoft, Oracle, Salesforce and SAP"
                                    }
                                ],
                            },
                        },
                    },
                ],
            },
        )
    ],
)
def test_convert_content_data_to_df_rdp(input_data):
    # given
    # when
    result = news_build_df_rdp(input_data)
    # then
    assert not result.empty


@pytest.mark.parametrize(
    ("input_data",),
    [
        (
            {
                "data": [],
            },
        )
    ],
)
def test_convert_content_data_to_df_rdp_empty_df(input_data):
    # given
    # when
    result = news_build_df_rdp(input_data)
    # then
    assert result.empty


def test__get_text_from_story():
    # given
    expected_result = "expected_result"
    input_story = {"newsItem": {"contentSet": {"inlineData": [{"$": expected_result}]}}}

    # when
    result = _get_text_from_story(input_story)

    # then
    assert result == expected_result


def test__get_headline_from_story():
    # given
    expected_result = "expected_result"
    input_story = {"newsItem": {"contentMeta": {"headline": [{"$": expected_result}]}}}

    # when
    result = _get_headline_from_story(input_story)

    # then
    assert result == expected_result



import json

import allure
import pytest

import refinitiv.data.content.ipa.financial_contracts as rdf
from refinitiv.data.errors import RDError
from tests.integration.constants_list import HttpStatusCode, HttpReason
from tests.integration.content.ipa.contracts.bond.conftest import (
    bond_definition_fixed_rate,
    bond_definition_floating_rate,
    bond_definition_user_defined,
    invalid_bond_definition,
    bond_universe,
)
from tests.integration.content.ipa.contracts.conftest import (
    check_http_status_is_success_and_df_value_not_empty,
    check_stream_state_and_df_from_stream,
)
from tests.integration.helpers import (
    check_response_status,
    get_async_response_from_definitions,
    check_dataframe_column_date_for_datetime_type,
)


@allure.suite("Content object - Bond")
@allure.feature("Content object - Bond")
@allure.severity(allure.severity_level.CRITICAL)
class TestBond:
    # EAPI-2456 bug when async_mode True in desktop session
    @allure.title("Create a bond definition object with params and get data")
    @pytest.mark.caseid("35996161")
    @pytest.mark.smoke
    @pytest.mark.parametrize(
        "bond_definition",
        [
            bond_definition_fixed_rate,
            bond_definition_floating_rate,
            bond_definition_user_defined,
        ],
    )
    def test_user_defined_bond(self, open_session, bond_definition):
        response = bond_definition().get_data()
        check_http_status_is_success_and_df_value_not_empty(response)
        check_dataframe_column_date_for_datetime_type(response)

    @allure.title("Create a bond definition request asynchronously")
    @pytest.mark.caseid("35996161")
    @pytest.mark.smoke
    async def test_bond_definition_asynchronously(self, open_session_async):
        valid_response, invalid_response = await get_async_response_from_definitions(
            bond_definition_fixed_rate(), invalid_bond_definition()
        )

        check_http_status_is_success_and_df_value_not_empty(valid_response)
        check_response_status(
            response=invalid_response,
            expected_status_code=HttpStatusCode.TWO_HUNDRED,
            expected_http_reason=HttpReason.OK,
        )

    @allure.title("Create a bond stream")
    @pytest.mark.caseid("36597161")
    def test_bond_stream_with_default_params(self, open_session):
        stream = bond_definition_fixed_rate().get_stream()
        stream.open()

        check_stream_state_and_df_from_stream(stream)

    @allure.title("Create a bond definition object with empty pricing parameters")
    @pytest.mark.parametrize("universe", ["US10YT=RR"])
    @pytest.mark.caseid("35984974")
    def test_multiple_bond_as_definitions_empty_pricing_parameters(
        self, universe, open_session
    ):
        response = rdf.bond.Definition(
            instrument_code=universe,
            pricing_parameters=rdf.bond.PricingParameters(),
        ).get_data()

        check_http_status_is_success_and_df_value_not_empty(response)

    @allure.title(
        "Create a bond definition object with different value type in universe"
    )
    @pytest.mark.parametrize("instrument_code", [2, ["1", "1"], {"1": "1"}])
    @pytest.mark.caseid("36051507")
    def test_bond_definition_with_different_value_type(
        self, instrument_code, open_platform_session
    ):
        with pytest.raises(ValueError) as error:
            rdf.bond.Definition(instrument_code=instrument_code).get_data()
        assert "Invalid type of instrument_code" in str(error.value)

    @allure.title("Create an invalid bond definition object and get RDError")
    @pytest.mark.caseid("")
    def test_invalid_bond_definition(self, open_session):
        with pytest.raises(RDError) as error:
            invalid_bond_definition().get_data()
        assert (
            str(error.value)
            == "Error code None | Market data error : There is no analytics for instrumentCode [ INVALID_INSTRUMENT_CODE ] in the data sourcing system that is used for pricing."
        )

    @allure.title("Create a bond definition with extended parameters")
    @pytest.mark.caseid("C37678339")
    def test_get_bond_with_extended_params(self, open_session):
        response = rdf.bond.Definition(
            issue_date="2002-02-28",
            end_date="2032-02-28",
            notional_ccy="USD",
            interest_payment_frequency="Annual",
            interest_calculation_method=rdf.bond.DayCountBasis.DCB_ACTUAL_ACTUAL,
            extended_params={
                "instrumentDefinition": {
                    "fixedRatePercent": 9,
                    "interestPaymentFrequency": "Annual",
                },
                "pricingParameters": {"cleanPrice": 122},
            },
        ).get_data()

        check_http_status_is_success_and_df_value_not_empty(response)

        request = json.loads(response.request_message.content.decode("utf-8"))
        request_universe = request.get("universe")[0]
        assert (
            request_universe == bond_universe
        ), f"Extended params are applied incorrectly"



from datetime import timedelta, datetime

import allure
import pytest

import refinitiv.data as rd
from refinitiv.data.content import news
from tests.integration.content.news.conftest import (
    check_data_in_response_sorted_by_order,
    check_dates_range_in_response,
)
from tests.integration.helpers import (
    check_index_column_contains_dates,
    check_dataframe_column_date_for_datetime_type,
    check_if_dataframe_is_not_none,
)


@allure.suite("FinCoder Layer")
@allure.feature("FinCoder - News Headlines")
@allure.severity(allure.severity_level.CRITICAL)
@pytest.mark.underlying_content("news")
class TestGetHeadlines:
    @allure.title("Get News Headlines with headline query")
    @pytest.mark.caseid("")
    @pytest.mark.parametrize(
        "query,default_count", [("(Britain and (Biden or Trump)) in 2022", 10)]
    )
    def test_get_news_headlines(self, open_session, query, default_count):
        response = rd.news.get_headlines(query=query)
        check_if_dataframe_is_not_none(response)
        check_index_column_contains_dates(response)
        check_dataframe_column_date_for_datetime_type(response)
        assert (
            response.shape[0] == default_count
        ), f"Actual Headlines amount is {response.shape[0]}"

    @allure.title("Get News Headlines with specified count param")
    @pytest.mark.caseid("")
    @pytest.mark.parametrize("sourceCode,count", [("NS:RTRS", 30)])
    def test_get_news_headlines_with_specified_count_param(
        self, open_session, sourceCode, count
    ):
        response = rd.news.get_headlines(query=sourceCode, count=count)
        check_if_dataframe_is_not_none(response)
        check_index_column_contains_dates(response)
        check_dataframe_column_date_for_datetime_type(response)
        assert response.shape[0] == count, f"Headlines amount is {response.shape[0]}"
        assert sourceCode in set(
            response.sourceCode.values
        ), "Inconsistency with result sourceCode"

    @allure.title("Get News Headlines with specified star and end parameters")
    @pytest.mark.caseid("")
    @pytest.mark.parametrize(
        "query,start,end,count",
        [
            (
                "P:4297144968",
                "2021-05-15T13:12:33.307Z",
                "2021-11-29T15:15:10.006Z",
                150,
            ),
            ("R:.FTSE", timedelta(hours=-10), timedelta(0), 5),
            ("Tesla", datetime(2022, 11, 10), datetime.now(), 100),
        ],
        ids=["string", "timedelta", "datetime"],
    )
    def test_get_news_headlines_with_specified_start_end_date(
        self,
        open_session,
        query,
        start,
        end,
        count,
    ):
        response = rd.news.get_headlines(query, start=start, end=end, count=count)

        check_if_dataframe_is_not_none(response)
        check_index_column_contains_dates(response)
        check_dataframe_column_date_for_datetime_type(response)
        check_dates_range_in_response(response, start, end)
        assert (
            response.shape[0] == count
        ), f"Actual Headlines amount is {response.shape[0]}"

    @allure.title("Get News Headlines with closed session")
    @pytest.mark.caseid("")
    @pytest.mark.parametrize("query", ["Microsoft"])
    def test_get_news_headlines_with_closed_session(self, query, open_session):
        session = open_session
        session.close()
        with pytest.raises(ValueError) as error:
            rd.news.get_headlines(query=query)
        assert str(error.value) == "Session is not opened. Can't send any request"

    @allure.title(
        "Get News headlines with specified sort_order param new-to-old/old-to-new "
    )
    @pytest.mark.caseid("")
    @pytest.mark.parametrize(
        "query,sort_order",
        [
            ("R:0#.FTSE", news.headlines.SortOrder.new_to_old),
            ("R:0#.FTSE", news.headlines.SortOrder.old_to_new),
        ],
    )
    def test_get_news_headlines_with_specified_sort_order_param(
        self, open_platform_session, query, sort_order
    ):
        response = rd.news.get_headlines(query=query, order_by=sort_order)
        check_if_dataframe_is_not_none(response)
        check_index_column_contains_dates(response)
        check_dataframe_column_date_for_datetime_type(response)
        check_data_in_response_sorted_by_order(response, sort_order)



from typing import TYPE_CHECKING

from ....content._content_response_factory import ContentResponseFactory

if TYPE_CHECKING:
    from ....delivery._data._parsed_data import ParsedData


class DatesAndCalendarsResponseFactory(ContentResponseFactory):
    def create_fail(self, parsed_data: "ParsedData", **kwargs):
        errors = parsed_data.status.get("error", {}).get("errors")
        if errors:
            parsed_data.error_messages = (
                f"{parsed_data.first_error_message}. {errors[0]['reason']}"
            )
        return super().create_fail(parsed_data, **kwargs)



import numpy as np
from typing import List, Optional, Union

from ..._types import OptDateTime
from ...content.ipa._enums import DateMovingConvention, EndOfMonthConvention
from ...content.ipa.dates_and_calendars.add_periods import Definition


def add_periods(
    start_date: "OptDateTime" = None,
    period: str = None,
    calendars: Optional[List[str]] = None,
    currencies: Optional[List[str]] = None,
    date_moving_convention: Optional[Union[DateMovingConvention, str]] = None,
    end_of_month_convention: Optional[Union[EndOfMonthConvention, str]] = None,
) -> np.datetime64:
    """
    With this function you can add a time period to a given date using a given calendar.

    Parameters
    ----------
        start_date: str or datetime or timedelta, optional
            Start date of calculation.
        period: str, optional
            String representing the tenor.
        calendars: list of str, optional
            Calendars to use the date for working day or weekend.
            Optional if currencies is provided.
        currencies: list of str, optional
            Currencies to use the date for working day or weekend.
            Optional if calendars is provided.
        date_moving_convention : DateMovingConvention or str, optional
            The method to adjust dates.
        end_of_month_convention : EndOfMonthConvention or str, optional
            End of month convention.

    Returns
    -------
    np.datetime64
        Added period date

    Examples
    --------
     >>> import datetime
     >>> import refinitiv.data as rd
     >>> from refinitiv.data import dates_and_calendars
     >>>
     >>> rd.open_session("platform.default")
     >>>
     >>> added_period = rd.dates_and_calendars.add_periods(
     ...    start_date=datetime.date(2014, 1, 1),
     ...    period="1Y",
     ...    calendars=["BAR", "KOR"],
     ...    date_moving_convention=dates_and_calendars.DateMovingConvention.NEXT_BUSINESS_DAY,
     ...    end_of_month_convention=dates_and_calendars.EndOfMonthConvention.LAST28
     ... )
    """

    response = Definition(
        start_date=start_date,
        period=period,
        calendars=calendars,
        currencies=currencies,
        date_moving_convention=date_moving_convention,
        end_of_month_convention=end_of_month_convention,
    ).get_data()

    return np.datetime64(response.data.added_period.date)



import abc
from datetime import timedelta
from typing import Callable, List, Tuple, Any, Optional, TYPE_CHECKING

from ._historical_join_responses import historical_join_responses
from .._content_type import ContentType
from .._tools import hp_datetime_adapter
from ..content._intervals import (
    interval_arg_parser,
    get_day_interval_type,
    DayIntervalType,
    Intervals,
)


if TYPE_CHECKING:
    from ..delivery._data._data_provider import Response


INTERVALS_BY_SECONDS = {
    Intervals.ONE_MINUTE.value: 59,
    Intervals.FIVE_MINUTES.value: 299,
    Intervals.TEN_MINUTES.value: 599,
    Intervals.THIRTY_MINUTES.value: 1799,
    Intervals.SIXTY_MINUTES.value: 3599,
    Intervals.HOURLY.value: 3599,
}

EVENTS_MAX_LIMIT = 10000


def remove_last_date_elements(data: List[List[Any]]) -> List[List[Any]]:
    end_date = data[-1][0]
    for index, item in enumerate(data[::-1]):
        if item[0] != end_date:
            data = data[:-index]
            return data

    return data


def create_raw(responses: List["Response"], entire_data: List[List[Any]]) -> dict:
    for response in responses:
        if response.is_success:
            raw = response.data.raw
            raw["data"] = entire_data
            return response.data.raw

    return {}


class EntireDataProvider(abc.ABC):
    @abc.abstractmethod
    def request_with_dates(self, *args) -> Tuple[List["Response"], List[List[Any]]]:
        pass

    @abc.abstractmethod
    def request_with_count(self, *args) -> Tuple[List["Response"], List[List[Any]]]:
        pass

    @abc.abstractmethod
    def get_request_function(self, **kwargs) -> Optional[Callable]:
        pass

    @abc.abstractmethod
    def get_request_function_async(self, **kwargs) -> Optional[Callable]:
        pass

    def get_data(self, provide_data: Callable, **kwargs) -> "Response":
        request_function = self.get_request_function(**kwargs)

        if request_function:
            responses, entire_data = request_function(provide_data, **kwargs)
            raw = create_raw(responses, entire_data)

        else:
            response = provide_data(**kwargs)
            responses = [response]
            raw = response.data.raw

        response = historical_join_responses(responses, raw)
        return response

    async def get_data_async(self, provide_data: Callable, **kwargs) -> "Response":
        request_function = self.get_request_function_async(**kwargs)

        if request_function:
            responses, entire_data = await request_function(provide_data, **kwargs)
            raw = create_raw(responses, entire_data)

        else:
            response = await provide_data(**kwargs)
            responses = [response]
            raw = response.data.raw

        response = historical_join_responses(responses, raw)
        return response


class SummariesEntireDataProvider(EntireDataProvider):
    def request_with_dates(
        self,
        provide_data: Callable,
        interval,
        start: str,
        end: str,
        count: Optional[int] = None,
        **kwargs,
    ) -> Tuple[List["Response"], List[List[Any]]]:
        interval_sec = INTERVALS_BY_SECONDS[interval_arg_parser.get_str(interval)]

        entire_data = []
        responses = []
        unique_data_count = set()

        finished_date = hp_datetime_adapter.get_localize(start)
        # need do ... while
        end_date = finished_date + timedelta(microseconds=1)

        while end_date > finished_date and len(unique_data_count) <= 1:
            response = provide_data(
                interval=interval,
                count=count,
                start=start,
                end=end,
                **kwargs,
            )
            responses.append(response)

            if not response.is_success:
                break

            raw = response.data.raw
            if len(raw) == 0 or not raw.get("data"):
                break

            data = raw["data"]
            entire_data.extend(data)

            if count is not None and len(entire_data) >= count:
                entire_data = entire_data[:count]
                break

            unique_data_count.add(len(data))

            end_date = data[-1][0]
            end = end_date
            end_date = hp_datetime_adapter.get_localize(end_date)

            if (end_date - finished_date).seconds < interval_sec:
                break

        return responses, entire_data

    async def request_with_dates_async(
        self,
        provide_data: Callable,
        interval,
        start: str,
        end: str,
        count: Optional[int] = None,
        **kwargs,
    ) -> Tuple[List["Response"], List[List[Any]]]:
        interval_sec = INTERVALS_BY_SECONDS[interval_arg_parser.get_str(interval)]

        entire_data = []
        responses = []
        unique_data_count = set()

        finished_date = hp_datetime_adapter.get_localize(start)
        # need do ... while
        end_date = finished_date + timedelta(microseconds=1)

        while end_date > finished_date and len(unique_data_count) <= 1:
            response = await provide_data(
                interval=interval,
                count=count,
                start=start,
                end=end,
                **kwargs,
            )
            responses.append(response)

            if not response.is_success:
                break

            raw = response.data.raw
            if len(raw) == 0 or not raw.get("data"):
                break

            data = raw["data"]
            entire_data.extend(data)

            if count is not None and len(entire_data) >= count:
                entire_data = entire_data[:count]
                break

            unique_data_count.add(len(data))

            end_date = data[-1][0]
            end = end_date
            end_date = hp_datetime_adapter.get_localize(end_date)

            if (end_date - finished_date).seconds < interval_sec:
                break

        return responses, entire_data

    def request_with_count(
        self,
        provide_data: Callable,
        interval,
        count: int,
        end: str,
        start: Optional[str] = None,
        **kwargs,
    ) -> Tuple[List["Response"], List[List[Any]]]:
        interval_sec = INTERVALS_BY_SECONDS[interval_arg_parser.get_str(interval)]

        c = count
        entire_data = []
        responses = []
        unique_data_count = set()

        finished_date = None
        if start:
            finished_date = hp_datetime_adapter.get_localize(start)

        while c > 0 and len(unique_data_count) <= 1:
            response = provide_data(
                interval=interval,
                count=count,
                start=start,
                end=end,
                **kwargs,
            )
            responses.append(response)

            if not response.is_success:
                break

            raw = response.data.raw
            if len(raw) == 0 or not raw.get("data"):
                break

            data = raw["data"]
            entire_data.extend(data)

            unique_data_count.add(len(data))

            c -= len(data)
            count = c
            end_date = data[-1][0]
            end = end_date

            if finished_date:
                end_date = hp_datetime_adapter.get_localize(end_date)

                if (end_date - finished_date).seconds < interval_sec:
                    break

        return responses, entire_data

    async def request_with_count_async(
        self,
        provide_data: Callable,
        interval,
        count: int,
        end: str,
        start: Optional[str] = None,
        **kwargs,
    ) -> Tuple[List["Response"], List[List[Any]]]:
        interval_sec = INTERVALS_BY_SECONDS[interval_arg_parser.get_str(interval)]

        c = count
        entire_data = []
        responses = []
        unique_data_count = set()

        finished_date = None
        if start:
            finished_date = hp_datetime_adapter.get_localize(start)

        while c > 0 and len(unique_data_count) <= 1:
            response = await provide_data(
                interval=interval,
                count=count,
                start=start,
                end=end,
                **kwargs,
            )
            responses.append(response)

            if not response.is_success:
                break

            raw = response.data.raw
            if len(raw) == 0 or not raw.get("data"):
                break

            data = raw["data"]
            entire_data.extend(data)

            unique_data_count.add(len(data))

            c -= len(data)
            count = c
            end_date = data[-1][0]
            end = end_date

            if finished_date:
                end_date = hp_datetime_adapter.get_localize(end_date)

                if (end_date - finished_date).seconds < interval_sec:
                    break

        return responses, entire_data

    def get_request_function(
        self,
        interval,
        count: Optional[int] = None,
        start: Optional[str] = None,
        end: Optional[str] = None,
        **kwargs,
    ) -> Optional[Callable]:
        request_function = None

        if (
            interval is None
            or get_day_interval_type(interval) is not DayIntervalType.INTRA
        ):
            return request_function

        if start is not None and end is not None:
            request_function = self.request_with_dates

        elif count is not None and count > 0:
            request_function = self.request_with_count

        return request_function

    def get_request_function_async(
        self,
        interval,
        count: Optional[int] = None,
        start: Optional[str] = None,
        end: Optional[str] = None,
        **kwargs,
    ) -> Optional[Callable]:
        request_function = None

        if (
            interval is None
            or get_day_interval_type(interval) is not DayIntervalType.INTRA
        ):
            return request_function

        if start is not None and end is not None:
            request_function = self.request_with_dates_async

        elif count is not None and count > 0:
            request_function = self.request_with_count_async

        return request_function


class EventsEntireDataProvider(EntireDataProvider):
    def request_with_dates(
        self,
        provide_data: Callable,
        start: str,
        end: str,
        count: Optional[int] = None,
        **kwargs,
    ) -> Tuple[List["Response"], List[List[Any]]]:
        entire_data = []
        responses = []

        finished_date = hp_datetime_adapter.get_localize(start)
        # need do ... while
        end_date = finished_date + timedelta(microseconds=1)
        response_count = EVENTS_MAX_LIMIT

        while end_date > finished_date and response_count >= EVENTS_MAX_LIMIT:
            response = provide_data(count=count, start=start, end=end, **kwargs)
            responses.append(response)

            if not response.is_success:
                break

            raw = response.data.raw
            if len(raw) == 0 or not raw.get("data"):
                break

            data = raw["data"]

            response_count = len(data)
            if response_count >= EVENTS_MAX_LIMIT:
                data = remove_last_date_elements(data)

            entire_data.extend(data)

            end_date = data[-1][0]
            end = end_date
            end_date = hp_datetime_adapter.get_localize(end_date)

            if count is not None and len(entire_data) >= count:
                entire_data = entire_data[:count]
                break

        return responses, entire_data

    async def request_with_dates_async(
        self,
        provide_data: Callable,
        start: str,
        end: str,
        count: Optional[int] = None,
        **kwargs,
    ) -> Tuple[List["Response"], List[List[Any]]]:
        entire_data = []
        responses = []

        finished_date = hp_datetime_adapter.get_localize(start)
        # need do ... while
        end_date = finished_date + timedelta(microseconds=1)
        response_count = EVENTS_MAX_LIMIT

        while end_date > finished_date and response_count >= EVENTS_MAX_LIMIT:
            response = await provide_data(count=count, start=start, end=end, **kwargs)
            responses.append(response)

            if not response.is_success:
                break

            raw = response.data.raw
            if len(raw) == 0 or not raw.get("data"):
                break

            data = raw["data"]

            response_count = len(data)
            if response_count >= EVENTS_MAX_LIMIT:
                data = remove_last_date_elements(data)

            entire_data.extend(data)

            end_date = data[-1][0]
            end = end_date
            end_date = hp_datetime_adapter.get_localize(end_date)

            if count is not None and len(entire_data) >= count:
                entire_data = entire_data[:count]
                break

        return responses, entire_data

    def request_with_count(
        self, provide_data: Callable, count: int, start: str, end: str, **kwargs
    ) -> Tuple[List["Response"], List[List[Any]]]:
        entire_data = []
        responses = []
        c = count
        response_count = EVENTS_MAX_LIMIT

        while c > 0 and response_count >= EVENTS_MAX_LIMIT:
            response = provide_data(count=count, start=start, end=end, **kwargs)
            responses.append(response)

            if not response.is_success:
                break

            raw = response.data.raw
            if len(raw) == 0 or not raw.get("data"):
                break

            data = raw["data"]

            response_count = len(data)
            if response_count >= EVENTS_MAX_LIMIT:
                data = remove_last_date_elements(data)

            entire_data.extend(data)

            c -= len(data)
            count = c
            end_date = data[-1][0]
            end = end_date

        return responses, entire_data

    async def request_with_count_async(
        self, provide_data: Callable, count: int, start: str, end: str, **kwargs
    ) -> Tuple[List["Response"], List[List[Any]]]:
        entire_data = []
        responses = []
        c = count
        response_count = EVENTS_MAX_LIMIT

        while c > 0 and response_count >= EVENTS_MAX_LIMIT:
            response = await provide_data(count=count, start=start, end=end, **kwargs)
            responses.append(response)

            if not response.is_success:
                break

            raw = response.data.raw
            if len(raw) == 0 or not raw.get("data"):
                break

            data = raw["data"]

            response_count = len(data)
            if response_count >= EVENTS_MAX_LIMIT:
                data = remove_last_date_elements(data)

            entire_data.extend(data)

            c -= len(data)
            count = c
            end_date = data[-1][0]
            end = end_date

        return responses, entire_data

    def get_request_function(
        self,
        count: Optional[int] = None,
        start: Optional[str] = None,
        end: Optional[str] = None,
        **kwargs,
    ) -> Optional[Callable]:
        request_function = None

        if start is not None and end is not None:
            request_function = self.request_with_dates

        elif count is not None and count > EVENTS_MAX_LIMIT:
            request_function = self.request_with_count

        return request_function

    def get_request_function_async(
        self,
        count: Optional[int] = None,
        start: Optional[str] = None,
        end: Optional[str] = None,
        **kwargs,
    ) -> Optional[Callable]:
        request_function = None

        if start is not None and end is not None:
            request_function = self.request_with_dates_async

        elif count is not None and count > EVENTS_MAX_LIMIT:
            request_function = self.request_with_count_async

        return request_function


entire_data_provider_by_content_type = {
    ContentType.HISTORICAL_PRICING_EVENTS: EventsEntireDataProvider(),
    ContentType.CUSTOM_INSTRUMENTS_EVENTS: EventsEntireDataProvider(),
    ContentType.HISTORICAL_PRICING_INTERDAY_SUMMARIES: SummariesEntireDataProvider(),
    ContentType.HISTORICAL_PRICING_INTRADAY_SUMMARIES: SummariesEntireDataProvider(),
    ContentType.CUSTOM_INSTRUMENTS_INTERDAY_SUMMARIES: SummariesEntireDataProvider(),
    ContentType.CUSTOM_INSTRUMENTS_INTRADAY_SUMMARIES: SummariesEntireDataProvider(),
}


def get_entire_data_provider(content_type: ContentType) -> EntireDataProvider:
    entire_data_provider = entire_data_provider_by_content_type.get(content_type)

    if not entire_data_provider:
        raise ValueError(f"Cannot find entire data provider for {content_type}")

    return entire_data_provider



import pytest

from refinitiv.data import OpenState
from refinitiv.data._core.session.event_code import EventCode
from ..conftest import create_platform_session, create_desktop_session

PLATFORM_SEQUENCE = [
    ("state", OpenState.Pending),
    ("event", EventCode.SessionAuthenticationSuccess),
    ("state", OpenState.Opened),
    ("state", OpenState.Closed),
]

DESKTOP_SEQUENCE = [
    ("state", OpenState.Pending),
    ("state", OpenState.Opened),
    ("state", OpenState.Closed),
]


@pytest.mark.parametrize(
    "session_provider, expected_sequence",
    [
        (create_platform_session, PLATFORM_SEQUENCE),
        (create_desktop_session, DESKTOP_SEQUENCE),
    ],
    ids=["platform_session", "desktop_session"],
)
def test_sessions_callbacks_sequence(session_provider, expected_sequence):
    session = session_provider()
    callbacks = []

    session.on_event(lambda *a: callbacks.append(("event", a[0])))
    session.on_state(lambda *a: callbacks.append(("state", a[0])))

    session.open()
    session.close()

    assert callbacks == expected_sequence, "Unexpected callbacks sequence"



# coding: utf8

from typing import Optional

from ..._object_definition import ObjectDefinition

from ._repo_parameters import RepoParameters
from ..bond import PricingParameters as BondPricingParameters


class UnderlyingPricingParameters(ObjectDefinition):
    """
    API endpoint for Financial Contract analytics,
    that returns calculations relevant to each contract type.

    Parameters
    ----------
    pricing_parameters_at_end : BondPricingParameters, optional

    pricing_parameters_at_start : BondPricingParameters, optional

    repo_parameters : RepoParameters, optional

    market_data_date : str, optional
        The date at which the market data is retrieved. the value is expressed in iso
        8601 format: yyyy-mm-ddt[hh]:[mm]:[ss]z (e.g., '2021-01-01t00:00:00z'). it
        should be less or equal tovaluationdate). optional. by
        default,marketdatadateisvaluationdateor today.
    report_ccy : str, optional
        The reporting currency code, expressed in iso 4217 alphabetical format (e.g.,
        'usd'). it is set for the fields ending with 'xxxinreportccy'. optional. the
        default value is the notional currency.
    valuation_date : str, optional
        The date at which the instrument is valued. the value is expressed in iso 8601
        format: yyyy-mm-ddt[hh]:[mm]:[ss]z (e.g., '2021-01-01t00:00:00z'). by default,
        marketdatadate is used. if marketdatadate is not specified, the default value is
        today.
    """

    def __init__(
        self,
        pricing_parameters_at_end: Optional[BondPricingParameters] = None,
        pricing_parameters_at_start: Optional[BondPricingParameters] = None,
        repo_parameters: Optional[RepoParameters] = None,
        market_data_date: Optional[str] = None,
        report_ccy: Optional[str] = None,
        valuation_date: Optional[str] = None,
    ) -> None:
        super().__init__()
        self.pricing_parameters_at_end = pricing_parameters_at_end
        self.pricing_parameters_at_start = pricing_parameters_at_start
        self.repo_parameters = repo_parameters
        self.market_data_date = market_data_date
        self.report_ccy = report_ccy
        self.valuation_date = valuation_date

    @property
    def pricing_parameters_at_end(self):
        """
        :return: object BondPricingParameters
        """
        return self._get_object_parameter(
            BondPricingParameters, "pricingParametersAtEnd"
        )

    @pricing_parameters_at_end.setter
    def pricing_parameters_at_end(self, value):
        self._set_object_parameter(
            BondPricingParameters, "pricingParametersAtEnd", value
        )

    @property
    def pricing_parameters_at_start(self):
        """
        :return: object BondPricingParameters
        """
        return self._get_object_parameter(
            BondPricingParameters, "pricingParametersAtStart"
        )

    @pricing_parameters_at_start.setter
    def pricing_parameters_at_start(self, value):
        self._set_object_parameter(
            BondPricingParameters, "pricingParametersAtStart", value
        )

    @property
    def repo_parameters(self):
        """
        :return: object RepoParameters
        """
        return self._get_object_parameter(RepoParameters, "repoParameters")

    @repo_parameters.setter
    def repo_parameters(self, value):
        self._set_object_parameter(RepoParameters, "repoParameters", value)

    @property
    def market_data_date(self):
        """
        The date at which the market data is retrieved. the value is expressed in iso
        8601 format: yyyy-mm-ddt[hh]:[mm]:[ss]z (e.g., '2021-01-01t00:00:00z'). it
        should be less or equal tovaluationdate). optional. by
        default,marketdatadateisvaluationdateor today.
        :return: str
        """
        return self._get_parameter("marketDataDate")

    @market_data_date.setter
    def market_data_date(self, value):
        self._set_parameter("marketDataDate", value)

    @property
    def report_ccy(self):
        """
        The reporting currency code, expressed in iso 4217 alphabetical format (e.g.,
        'usd'). it is set for the fields ending with 'xxxinreportccy'. optional. the
        default value is the notional currency.
        :return: str
        """
        return self._get_parameter("reportCcy")

    @report_ccy.setter
    def report_ccy(self, value):
        self._set_parameter("reportCcy", value)

    @property
    def valuation_date(self):
        """
        The date at which the instrument is valued. the value is expressed in iso 8601
        format: yyyy-mm-ddt[hh]:[mm]:[ss]z (e.g., '2021-01-01t00:00:00z'). by default,
        marketdatadate is used. if marketdatadate is not specified, the default value is
        today.
        :return: str
        """
        return self._get_parameter("valuationDate")

    @valuation_date.setter
    def valuation_date(self, value):
        self._set_parameter("valuationDate", value)



from typing import TYPE_CHECKING

from ..._content_data import Data
from ..._content_provider_layer import ContentUsageLoggerMixin
from ...._content_type import ContentType
from ...._tools import validate_types, validate_bool_value, try_copy_to_list
from ....delivery._data._data_provider import DataProviderLayer, BaseResponse

if TYPE_CHECKING:
    from ...._types import ExtendedParams, StrStrings


class Definition(
    ContentUsageLoggerMixin[BaseResponse[Data]],
    DataProviderLayer[BaseResponse[Data]],
):
    """
    This class describe parameters to retrieve the calculated concentration data
    by top 10, 20, 50, 100 fund investors.

    Parameters
    ----------
    universe: str, list of str
        The Universe parameter allows the user to define the companies for which the content is returned.

    count: int
        Number of records

    use_field_names_in_headers: bool, optional
        Return field name as column headers for data instead of title

    extended_params : ExtendedParams, optional
        If necessary other parameters.

    Examples
    --------
    >>> from refinitiv.data.content import ownership
    >>> definition = ownership.fund.top_n_concentration.Definition("TRI.N", 30)
    >>> response = definition.get_data()
    """

    _USAGE_CLS_NAME = "Ownership.Fund.TopNConcentrationDefinition"

    def __init__(
        self,
        universe: "StrStrings",
        count: int,
        use_field_names_in_headers: bool = False,
        extended_params: "ExtendedParams" = None,
    ):
        validate_types(count, [int], "count")
        validate_bool_value(use_field_names_in_headers)
        universe = try_copy_to_list(universe)

        super().__init__(
            ContentType.OWNERSHIP_FUND_TOP_N_CONCENTRATION,
            universe=universe,
            count=count,
            use_field_names_in_headers=use_field_names_in_headers,
            extended_params=extended_params,
        )



__all__ = (
    "search",
    "Views",
    "search_templates",
    "convert_symbols",
    "Peers",
    "Chain",
    "Screener",
    "AssetClass",
    "AssetState",
    "CountryCode",
    "SymbolTypes",
)

from ._search import search
from ..content.search import Views
from ..content.symbol_conversion import AssetClass, AssetState, CountryCode, SymbolTypes
from ._search_templates import templates as search_templates
from ._universe_expanders import Peers, Chain, Screener
from ._convert_symbols import convert_symbols



from ..._enums import (
    Axis,
    DiscountingType,
    EtiInputVolatilityType,
    Format,
    FxSwapCalculationMethod,
    FxVolatilityModel,
    InputVolatilityType,
    PriceSide,
    TimeStamp,
    UnderlyingType,
    VolatilityAdjustmentType,
    VolatilityModel,
)
from ._outputs import *
from ._moneyness_type import *



from unittest import mock
import pytest

from refinitiv.data.content.esg._esg_data_provider import ErrorParser
from refinitiv.data.delivery._data._data_provider import ParsedData


@pytest.mark.parametrize(
    ("input_parsed_data", "test_result"),
    [
        (
            ParsedData({}, {}),
            {
                "status": {},
                "raw_response": {},
                "content_data": None,
                "error_codes": [],
                "error_messages": [],
            },
        ),
        (
            ParsedData(
                {"error": {"errors": [{"reason": "test reason"}]}},
                {},
                error_messages=["wrong message"],
            ),
            {
                "status": {"error": {"errors": [{"reason": "test reason"}]}},
                "raw_response": {},
                "content_data": None,
                "error_codes": [0],
                "error_messages": ["wrong message: test reason"],
            },
        ),
    ],
)
def test_esg_parser_process_failed_response(input_parsed_data, test_result):
    # given
    # when
    with mock.patch(
        "refinitiv.data.delivery._data._data_provider.Parser.process_failed_response"
    ) as mock_func:
        mock_func.return_value = input_parsed_data
        result = ErrorParser().process_failed_response(...)

    # then
    assert result == test_result



import pytest

from refinitiv.data.content.search import Views as SearchViews
from refinitiv.data.discovery._search_templates.manage import (
    depascalize_view,
    BuiltinNamespace,
)


def test_depascalize_view():
    for upper_value, enum_item in SearchViews.__members__.items():
        assert depascalize_view(enum_item.value) == upper_value


class TestBuiltinNamespace:
    def test_getitem(self):
        ns = BuiltinNamespace()
        assert ns["RICCategory"]

    def test_getitem_not_template(self):
        ns = BuiltinNamespace()
        with pytest.raises(AttributeError):
            ns["keys"]



import pytest

from refinitiv.data._content_type import ContentType
from refinitiv.data.delivery._data._data_provider_factory import make_provider
from refinitiv.data.delivery._data._data_type import DataType
from tests.unit.conftest import StubSession


@pytest.mark.parametrize(
    "content_type",
    [
        DataType.ENDPOINT,
        ContentType.NEWS_STORY_RDP,
        ContentType.NEWS_STORY_UDF,
        ContentType.NEWS_IMAGES,
        ContentType.DEFAULT,
        ContentType.FILINGS_SEARCH,
    ],
)
def test_response_data_does_not_have_df_property_simple_case(content_type):
    # given
    session = StubSession(is_open=True)

    # when
    response = make_provider(content_type).get_data(session, "url")

    # then
    assert hasattr(response.data, "raw")
    assert not hasattr(response.data, "df")



# coding: utf8

__all__ = ["PayoutScaling"]

from typing import Optional

from .._object_definition import ObjectDefinition


class PayoutScaling(ObjectDefinition):
    """
    Parameters
    ----------
    maximum : float, optional

    minimum : float, optional

    """

    def __init__(
        self,
        maximum: Optional[float] = None,
        minimum: Optional[float] = None,
    ) -> None:
        super().__init__()
        self.maximum = maximum
        self.minimum = minimum

    @property
    def maximum(self):
        """
        :return: float
        """
        return self._get_parameter("maximum")

    @maximum.setter
    def maximum(self, value):
        self._set_parameter("maximum", value)

    @property
    def minimum(self):
        """
        :return: float
        """
        return self._get_parameter("minimum")

    @minimum.setter
    def minimum(self, value):
        self._set_parameter("minimum", value)



import allure
import pytest

from refinitiv.data.content.ownership.insider import transaction_report
from tests.integration.constants_list import HttpStatusCode, HttpReason
from tests.integration.content.ownership.helpers import (
    check_response_dataframe_instrument_contains_expected_universes,
    get_expected_column,
)
from tests.integration.helpers import (
    get_async_response_from_definitions,
    check_response_status,
    check_non_empty_response_data,
    check_response_dataframe_contains_columns_names,
    check_extended_params_were_sent_in_request,
    check_dataframe_column_date_for_datetime_type,
)

EXPECTED_COLUMN_NAMES = [
    "instrument",
    "investorname",
    "date",
    "TR.InsiderInvestorType",
    "TR.InsiderRole",
    "TR.InsiderPrimaryRole",
    "TR.InsiderDate",
    "TR.InsiderIsNLE",
    "TR.InsiderIsNLEDate",
    "TR.InsiderInactiveDate",
    "TR.InvTrnIsNoLongerFiling",
    "TR.InvTransIsNLFDate",
    "TR.TransactionDate",
    "TR.TransactionTypeCode",
    "TR.TransactionType",
    "TR.AcquisitionType",
    "TR.AdjSharesTraded",
    "TR.AsRepSharesTraded",
    "TR.PctOfSharesOutTraded",
    "TR.TransactionPrice",
    "TR.InvTransPriceInARCurr",
    "TR.InvTransFilingType",
    "TR.TableNumber",
    "TR.IntradayTradeSeqNum",
    "TR.RoleSequenceNumber",
    "TR.RoleType",
    "TR.AmendmentId",
    "TR.InvestorTransDCN",
    "TR.TransactionCreateDate",
    "TR.TransactionUpdateDate",
    "TR.InvTransHoldingType",
    "TR.LatestHoldingType",
    "TR.AdjSharesHeld",
    "TR.AdjIndirectSharesHeld",
    "TR.AsRepSharesHeld",
    "TR.AsRepIndSharesHeld",
    "TR.InsiderRegion",
    "TR.InsAddressLine1",
    "TR.InsAddressLine2",
    "TR.InsAddressLine3",
    "TR.InsStateProv",
    "TR.InsAddressCity",
    "TR.InsCountry",
    "TR.InsPostalCode",
]

EXPECTED_COLUMN_TITLES = [
    "Instrument",
    "Equity owner name",
    "Date",
    "Investor Type",
    "Insider Role Type",
    "isPrimary InsiderRole",
    "Date Became Insider",
    "Contact NoLonger Exists",
    "Contact NoLonger Exists Date",
    "Insider Inactive Date",
    "Investment Turnsaction NoLongerFiles",
    "Date Considered NoLongerFiling",
    "Transaction Date",
    "Transaction TypeCode",
    "Transaction Type",
    "AcquisitionType",
    "Adjusted SharesTraded",
    "AsReported SharesTraded",
    "% Of SharesOutstandingTraded",
    "Transaction Price",
    "Transaction Price As Reported Curr",
    "Transaction Filing Type",
    "US Insider Form Table Num",
    "Intraday Trade SeqNumber",
    "Role SeqNum",
    "Title",
    "Amendment Indicator",
    "Filing DocControlNum",
    "Transaction CreateDate",
    "Transaction UpdateDate",
    "HoldingType",
    "Latest HoldingType",
    "Adjusted SharesHeld",
    "Adjusted Indirect SharesHeld",
    "AsReported SharesHeld",
    "AsReported Indirect SharesHeld",
    "Region",
    "AddrLine1",
    "AddrLine2",
    "AddrLine3",
    "State Province",
    "City",
    "Country",
    "PostalCode",
]


@allure.suite("Content object - Ownership Insider Transaction Report")
@allure.feature("Content object - Ownership Insider Transaction Report")
@allure.severity(allure.severity_level.CRITICAL)
class TestInsiderTransactionReport:
    @allure.title(
        "Create Insider Transaction Report definition definition object with valid required params and get data"
    )
    @pytest.mark.caseid("37556989")
    @pytest.mark.parametrize(
        "universe,use_field_names_in_headers,extended_params",
        [
            ("AAPL.O", True, {"test_param": "test_value"}),
            (["TRI.N", "ONEX.CCP", "AAPL.O"], False, {"test_param": "test_value"}),
            (["invalid universe", "ONEX.CCP"], True, None),
        ],
    )
    def test_create_insider_transaction_report_object_with_valid_fields_and_get_data(
        self,
        open_platform_session,
        universe,
        use_field_names_in_headers,
        extended_params,
    ):
        expected_column = get_expected_column(
            use_field_names_in_headers, EXPECTED_COLUMN_TITLES, EXPECTED_COLUMN_NAMES
        )
        response = transaction_report.Definition(
            universe=universe,
            use_field_names_in_headers=use_field_names_in_headers,
            extended_params=extended_params,
        ).get_data()
        print(response.data.df.to_string())
        check_response_status(response, HttpStatusCode.TWO_HUNDRED, HttpReason.OK)
        check_non_empty_response_data(response)
        check_response_dataframe_contains_columns_names(response, expected_column)
        check_dataframe_column_date_for_datetime_type(response)
        check_extended_params_were_sent_in_request(response, extended_params)
        # check_response_dataframe_instrument_contains_expected_universes(response, universe)
        #  - this check does not work yet while paging is not implemented, as all data returned
        #  for only the first universe, too many records. Uncomment with paging available.

    @allure.title(
        "Create Insider Transaction Report definition with title in headers by use_field_names_in_headers False as default"
    )
    @pytest.mark.caseid("C40219573")
    @pytest.mark.parametrize("universe", ["TRI.N"])
    def test_create_insider_transaction_report_object_with_use_field_names_in_headers_false(
        self, open_platform_session, universe
    ):
        response = transaction_report.Definition(
            universe, use_field_names_in_headers=False
        ).get_data()
        check_response_status(response, HttpStatusCode.TWO_HUNDRED, HttpReason.OK)
        check_non_empty_response_data(response)
        check_response_dataframe_contains_columns_names(
            response, EXPECTED_COLUMN_TITLES
        )
        check_dataframe_column_date_for_datetime_type(response)

    @allure.title(
        "Create Insider Transaction Report definition object with mix valid and invalid params and get data asynchronously"
    )
    @pytest.mark.caseid("37556990")
    @pytest.mark.parametrize(
        "universe,invalid_universe,start,end,limit",
        [
            (
                ["ONEX.CCP", "invalid universe"],
                "invalid universe",
                "20210725",
                "20211005",
                4,
            ),
        ],
    )
    @pytest.mark.smoke
    async def test_create_insider_transaction_report_object_with_mix_valid_and_invalid_params_and_get_data_async(
        self, open_platform_session_async, universe, invalid_universe, start, end, limit
    ):
        valid_response, invalid_response = await get_async_response_from_definitions(
            transaction_report.Definition(
                universe, start, end, limit, use_field_names_in_headers=True
            ),
            transaction_report.Definition(
                invalid_universe, start, end, limit, use_field_names_in_headers=True
            ),
        )
        check_response_status(
            response=valid_response,
            expected_status_code=HttpStatusCode.TWO_HUNDRED,
            expected_http_reason=HttpReason.OK,
            expected_error_code=None,
            expected_error_message="Failed to resolve identifiers ['invalid universe']",
        )
        check_non_empty_response_data(valid_response)
        check_response_status(
            response=invalid_response,
            expected_status_code=HttpStatusCode.TWO_HUNDRED,
            expected_http_reason=HttpReason.OK,
            expected_error_code=HttpStatusCode.FOUR_HUNDRED_TWELVE,
            expected_error_message="Unable to resolve all requested identifiers. Requested items: ['invalid universe']",
        )
        check_response_dataframe_contains_columns_names(
            valid_response, EXPECTED_COLUMN_NAMES
        )
        check_response_dataframe_instrument_contains_expected_universes(
            valid_response, universe[0]
        )



# coding: utf8

from typing import Optional

from .._base import UnderlyingDefinition


class EtiUnderlyingDefinition(UnderlyingDefinition):
    """
    Parameters
    ----------
    instrument_code : str, optional
        The underlier RIC
    """

    def __init__(
        self,
        instrument_code: Optional[str] = None,
    ) -> None:
        super().__init__()
        self.instrument_code = instrument_code

    @property
    def instrument_code(self):
        """
        The underlier RIC
        :return: str
        """
        return self._get_parameter("instrumentCode")

    @instrument_code.setter
    def instrument_code(self, value):
        self._set_parameter("instrumentCode", value)



import abc
from typing import Union


class ContribResponse(abc.ABC):
    def __init__(self, message: dict):
        self._message = message

    @property
    @abc.abstractmethod
    def is_success(self) -> bool:
        pass

    @property
    def type(self) -> str:
        return self._message.get("Type", "")

    @property
    def ack_id(self) -> Union[int, None]:
        return self._message.get("AckID")

    @property
    def nak_code(self) -> str:
        return self._message.get("NakCode", "")

    @property
    @abc.abstractmethod
    def nak_message(self) -> str:
        pass

    @property
    @abc.abstractmethod
    def error(self) -> str:
        pass

    @property
    def debug(self) -> dict:
        return self._message.get("Debug", {})


class AckContribResponse(ContribResponse):
    def __repr__(self) -> str:
        d = {"Type": self.type, "AckId": self.ack_id}

        if self.nak_code:
            d["NakCode"] = self.nak_code
            d["Message"] = self.nak_message

        return str(d)

    @property
    def is_success(self) -> bool:
        return not self.nak_code

    @property
    def nak_message(self) -> str:
        return self._message.get("Text", "")

    @property
    def error(self) -> str:
        return ""


class ErrorContribResponse(ContribResponse):
    def __repr__(self) -> str:
        d = {
            "Type": self.type,
            "Text": self.nak_message,
            "Debug": self.debug,
        }
        return str(d)

    @property
    def is_success(self) -> bool:
        return False

    @property
    def nak_message(self) -> str:
        return ""

    @property
    def error(self) -> str:
        return self._message.get("Text", "")


class NullContribResponse(ContribResponse):
    def __init__(self):
        super().__init__(None)

    @property
    def is_success(self) -> bool:
        return False

    @property
    def type(self) -> str:
        return ""

    @property
    def ack_id(self) -> Union[int, None]:
        return None

    @property
    def nak_code(self) -> str:
        return ""

    @property
    def nak_message(self) -> str:
        return ""

    @property
    def error(self) -> str:
        return ""

    @property
    def debug(self) -> dict:
        return {}



import allure
import pytest

import refinitiv.data as rd
from refinitiv.data.errors import RDError
from refinitiv.data.discovery import Peers, Screener
from tests.integration.access.conftest import (
    check_column_names_is_exist_in_response_and_df_not_empty,
)
from tests.integration.discovery.universe_expanders.conftest import (
    check_chain_discovery_object_for_attributes,
    is_iterable,
    FCHI_CONSTITUENTS,
    FCHI_SUMMARY_LINKS,
    LSEG_PEERS_INSTRUMENTS,
)
from tests.integration.helpers import compare_list


@allure.suite("Discovery layer")
@allure.feature("Discovery layer - Universe Expander")
@allure.severity(allure.severity_level.CRITICAL)
@pytest.mark.underlying_content("datagrid")
class TestExpanders:
    @allure.title(
        "Create discovery chain object depend on scopes with different endpoints"
    )
    @pytest.mark.parametrize(
        "chain_name,scope,account",
        [
            ("0#.FCHI", "trapi.streaming.pricing.read", "machine"),
            (
                "0#.FCHI",
                "trapi.data.get.data.read",
                "eikon",
            ),
            pytest.param(
                "0#.FCHI",
                "trapi.data.pricing.read",
                "machine",
                marks=pytest.mark.skip(
                    reason="https://jira.refinitiv.com/browse/EAPI-4919"
                ),
            ),
        ],
        ids=[
            "get_chain_data_from_stream",
            "get_chain_data_from_adc",
            "get_chain_data_from_chain_endpoint",
        ],
    )
    @pytest.mark.caseid("C52511279")
    @pytest.mark.smoke
    def test_create_discovery_chain_object_depend_on_scopes_with_different_endpoints(
        self,
        request,
        create_session_with_scope,
        chain_name,
        scope,
        account,
    ):
        create_session_with_scope(scope=scope, acc=account)
        chain = rd.discovery.Chain(name=chain_name)
        list(chain.constituents)

        assert is_iterable(chain), "Chain object is not iterable"
        check_chain_discovery_object_for_attributes(
            chain, chain_name, FCHI_CONSTITUENTS
        )

        if "get_chain_data_from_stream" in request.node.callspec.id:
            compare_list(chain.summary_links, FCHI_SUMMARY_LINKS)

    @allure.title("Create discovery peers/screener object and get mix of pricing data")
    @pytest.mark.parametrize(
        "function,expression,fields,expected_universe",
        [
            (Peers, "LSEG.L", None, LSEG_PEERS_INSTRUMENTS),
            (
                Screener,
                'U(IN(Equity(active,public,primary))/*UNV:Public*/), IN(TR.HQCountryCode,"AR"), IN(TR.GICSIndustryCode,"401010")',
                [
                    "TR.CommonName",
                    "TR.HeadquartersCountry",
                    "TR.GICSSector",
                    "TR.OrganizationStatusCode",
                    "TR.Revenue",
                    "BID",
                ],
                ["BMA.BA", "BBAR.BA", "BHIP.BA", "GGAL.BA", "BPAT.BA", "SUPV.BA"],
            ),
        ],
    )
    @pytest.mark.caseid("C52511280")
    @pytest.mark.smoke
    def test_create_discovery_peers_screeners_objects_and_get_data(
        self,
        set_underlying_platform_config,
        open_desktop_session,
        function,
        expression,
        fields,
        expected_universe,
    ):
        discovery_obj = function(expression=expression)
        actual_expression = discovery_obj.expression

        assert is_iterable(discovery_obj), f"{function} object is not iterable"
        assert (
            expression in actual_expression
        ), f"{expression} is missing for {function} object"
        assert (discovery_obj._universe, expected_universe)

        response = rd.get_data(
            universe=discovery_obj,
            fields=fields,
        )
        list_of_instruments = list(response["Instrument"])
        assert (list_of_instruments, expected_universe)
        check_column_names_is_exist_in_response_and_df_not_empty(response)

    @allure.title("Create discovery chain object with invalid name and get error")
    @pytest.mark.caseid("C52511281")
    @pytest.mark.parametrize("chain_name", ["invalid"])
    def test_create_chain_object_with_invalid_name_and_get_error(
        self, open_desktop_session, chain_name
    ):
        with pytest.raises(RDError) as error:
            chain = rd.discovery.Chain(name=chain_name)
            list(chain.constituents)
        assert str(error.value) == "Error code -1 | No values to unpack"

    @allure.title(
        "Create discovery peers/screeners object with invalid name and get error"
    )
    @pytest.mark.caseid("C52511297")
    @pytest.mark.parametrize(
        "function,error_message",
        [
            (
                Peers,
                "Error code 413 | Unable to resolve some identifier(s). Requested universes: peers(invalid). Requested fields: TR.RIC",
            ),
            (
                Screener,
                "Error code 800 | SCREEN(INVALID) processing failed. Requested universes: screen(invalid). Requested fields: TR.RIC",
            ),
        ],
    )
    def test_create_discovery_object_with_invalid_name_and_get_error(
        self, open_desktop_session, function, error_message
    ):
        with pytest.raises(RDError) as error:
            discovery_obj = function(expression="invalid")
            list(discovery_obj)
        assert str(error.value) == error_message



# coding: utf8

from typing import Optional

from ._enums import BuySell, CallPut, ExerciseStyle, UnderlyingType, SettlementType
from ._eti import (
    EtiUnderlyingDefinition,
    EtiBinaryDefinition,
    EtiBarrierDefinition,
    EtiDefinition,
    EtiCbbcDefinition,
    EtiDoubleBarriersDefinition,
    EtiFixingInfo,
)
from ._fx import (
    FxDefinition,
    FxDualCurrencyDefinition,
    FxDoubleBarrierDefinition,
    FxDoubleBinaryDefinition,
    FxForwardStart,
)
from ._option_definition import OptionDefinition
from ._models import InputFlow


class OptionInstrumentDefinition(EtiDefinition, FxDefinition, OptionDefinition):
    def __init__(
        self,
        asian_definition: Optional[EtiFixingInfo] = None,
        barrier_definition: Optional[EtiBarrierDefinition] = None,
        binary_definition: Optional[EtiBinaryDefinition] = None,
        buy_sell: Optional[BuySell] = None,
        call_put: Optional[CallPut] = None,
        cbbc_definition: Optional[EtiCbbcDefinition] = None,
        deal_contract: Optional[int] = None,
        delivery_date: Optional[str] = None,
        double_barrier_definition: Optional[FxDoubleBarrierDefinition] = None,
        double_barriers_definition: Optional[EtiDoubleBarriersDefinition] = None,
        double_binary_definition: Optional[FxDoubleBinaryDefinition] = None,
        dual_currency_definition: Optional[FxDualCurrencyDefinition] = None,
        end_date: Optional[str] = None,
        end_date_time: Optional[str] = None,
        exercise_style: Optional[ExerciseStyle] = None,
        forward_start_definition: Optional[FxForwardStart] = None,
        instrument_code: Optional[str] = None,
        instrument_tag: Optional[str] = None,
        lot_size: Optional[float] = None,
        notional_amount: Optional[float] = None,
        notional_ccy: Optional[str] = None,
        payments: Optional[InputFlow] = None,
        settlement_ccy: Optional[str] = None,
        settlement_type: Optional[SettlementType] = None,
        start_date: Optional[str] = None,
        strike: Optional[float] = None,
        tenor: Optional[str] = None,
        time_zone_offset: Optional[int] = None,
        underlying_definition: Optional[EtiUnderlyingDefinition] = None,
        underlying_type: Optional[UnderlyingType] = None,
    ):
        super().__init__(
            asian_definition=asian_definition,
            barrier_definition=barrier_definition,
            binary_definition=binary_definition,
            buy_sell=buy_sell,
            call_put=call_put,
            cbbc_definition=cbbc_definition,
            deal_contract=deal_contract,
            delivery_date=delivery_date,
            double_barrier_definition=double_barrier_definition,
            double_barriers_definition=double_barriers_definition,
            double_binary_definition=double_binary_definition,
            dual_currency_definition=dual_currency_definition,
            end_date=end_date,
            end_date_time=end_date_time,
            exercise_style=exercise_style,
            forward_start_definition=forward_start_definition,
            instrument_code=instrument_code,
            instrument_tag=instrument_tag,
            lot_size=lot_size,
            notional_amount=notional_amount,
            notional_ccy=notional_ccy,
            payments=payments,
            settlement_ccy=settlement_ccy,
            settlement_type=settlement_type,
            start_date=start_date,
            strike=strike,
            tenor=tenor,
            time_zone_offset=time_zone_offset,
            underlying_definition=underlying_definition,
            underlying_type=underlying_type,
        )



from typing import TYPE_CHECKING

from ._base_definition import BaseDefinition
from ..._content_type import ContentType
from ..._tools import validate_types, validate_bool_value, try_copy_to_list

if TYPE_CHECKING:
    from ..._types import StrStrings, OptInt


class Definition(BaseDefinition):
    """
    This class describe parameters to retrieve ESG standart measures data.

    Parameters
    ----------
    universe : str, list of str
        The Universe parameter allows the user to define the company they
        want content returned for, ESG content is delivered at the Company Level.

    start : int, optional
        Start & End parameter allows the user to request
         the number of Financial Years they would like returned.

    end : int, optional
        Start & End parameter allows the user to request
        the number of Financial Years they would like returned.

    use_field_names_in_headers: bool, optional
        Return field name as column headers for data instead of title

    Examples
    --------
    >>> from refinitiv.data.content import esg
    >>> definition = esg.standard_measures.Definition("BNPP.PA")
    """

    def __init__(
        self,
        universe: "StrStrings",
        start: "OptInt" = None,
        end: "OptInt" = None,
        use_field_names_in_headers: bool = False,
    ):
        validate_types(start, [int, type(None)], "start")
        validate_types(end, [int, type(None)], "end")
        validate_bool_value(use_field_names_in_headers)
        universe = try_copy_to_list(universe)

        super().__init__(
            ContentType.ESG_STANDARD_MEASURES,
            universe=universe,
            start=start,
            end=end,
            use_field_names_in_headers=use_field_names_in_headers,
        )

    def __repr__(self):
        get_repr = super().__repr__()
        return get_repr.replace("esg", "esg.standard_measures")



from refinitiv.data.content.ipa.curves._cross_currency_curves import curves
from tests.unit.conftest import StubResponse, StubSession


def test_definition():
    # given
    response = StubResponse(
        {
            "data": [
                {
                    "curveDefinition": {
                        "baseCurrency": "EUR",
                        "baseIndexName": "ESTR",
                        "quotedCurrency": "USD",
                        "quotedIndexName": "SOFR",
                        "crossCurrencyDefinitions": [
                            {
                                "baseCurrency": "EUR",
                                "baseIndexName": "ESTR",
                                "name": "EUR ESTR/USD SOFR FxCross",
                                "quotedCurrency": "USD",
                                "quotedIndexName": "SOFR",
                                "source": "Refinitiv",
                                "isNonDeliverable": False,
                                "mainConstituentAssetClass": "Swap",
                                "riskType": "CrossCurrency",
                                "id": "c9f2e9fb-b04b-4140-8377-8b7e47391486",
                                "ignoreExistingDefinition": False,
                            }
                        ],
                    },
                    "curveParameters": {
                        "valuationDate": "2021-10-06",
                        "interpolationMode": "Linear",
                        "extrapolationMode": "Constant",
                        "turnAdjustments": {},
                        "ignorePivotCurrencyHolidays": False,
                        "useDelayedDataIfDenied": False,
                        "ignoreInvalidInstrument": True,
                        "marketDataLookBack": {"value": 10, "unit": "CalendarDay"},
                    },
                    "curve": {
                        "fxCrossScalingFactor": 1.0,
                        "fxSwapPointScalingFactor": 10000.0,
                        "curvePoints": [
                            {
                                "tenor": "SPOT",
                                "startDate": "2021-10-08",
                                "endDate": "2021-10-08",
                                "swapPoint": {"bid": 0.0, "ask": 0.0, "mid": 0.0},
                                "outright": {
                                    "bid": 1.1556,
                                    "ask": 1.156,
                                    "mid": 1.1558,
                                },
                            },
                            {
                                "tenor": "9M",
                                "startDate": "2021-10-08",
                                "endDate": "2022-07-08",
                                "instruments": [{"instrumentCode": "EUR9M="}],
                                "swapPoint": {
                                    "bid": 66.25000000000102,
                                    "ask": 69.42999999999921,
                                    "mid": 67.84000000000012,
                                },
                                "outright": {
                                    "bid": 1.162225,
                                    "ask": 1.1629429999999998,
                                    "mid": 1.1625839999999998,
                                },
                            },
                        ],
                    },
                }
            ]
        }
    )
    session = StubSession(is_open=True, response=response)
    definition = curves.Definition(
        curve_definition=curves.FxForwardCurveDefinition(
            base_currency="EUR",
            base_index_name="ESTR",
            quoted_currency="USD",
            quoted_index_name="SOFR",
        ),
        curve_parameters=curves.FxForwardCurveParameters(valuation_date="2021-10-06"),
    )

    # when
    response = definition.get_data(session)

    # then
    assert response.data.raw
    assert not response.data.df.empty



# coding: utf8
from typing import Optional

from .._enums import BarrierType
from .._object_definition import ObjectDefinition


class BarrierDefinitionElement(ObjectDefinition):
    def __init__(
        self,
        barrier_type: Optional[BarrierType] = None,
        barrier_down_percent: Optional[float] = None,
        barrier_up_percent: Optional[float] = None,
        rebate_down_percent: Optional[float] = None,
        rebate_up_percent: Optional[float] = None,
    ):
        super().__init__()
        self.barrier_type = barrier_type
        self.barrier_down_percent = barrier_down_percent
        self.barrier_up_percent = barrier_up_percent
        self.rebate_down_percent = rebate_down_percent
        self.rebate_up_percent = rebate_up_percent

    @property
    def barrier_type(self):
        """
        :return: enum BarrierType
        """
        return self._get_enum_parameter(BarrierType, "barrierType")

    @barrier_type.setter
    def barrier_type(self, value):
        self._set_enum_parameter(BarrierType, "barrierType", value)

    @property
    def barrier_down_percent(self):
        """
        :return: float
        """
        return self._get_parameter("barrierDownPercent")

    @barrier_down_percent.setter
    def barrier_down_percent(self, value):
        self._set_parameter("barrierDownPercent", value)

    @property
    def barrier_up_percent(self):
        """
        :return: float
        """
        return self._get_parameter("barrierUpPercent")

    @barrier_up_percent.setter
    def barrier_up_percent(self, value):
        self._set_parameter("barrierUpPercent", value)

    @property
    def rebate_down_percent(self):
        """
        :return: float
        """
        return self._get_parameter("rebateDownPercent")

    @rebate_down_percent.setter
    def rebate_down_percent(self, value):
        self._set_parameter("rebateDownPercent", value)

    @property
    def rebate_up_percent(self):
        """
        :return: float
        """
        return self._get_parameter("rebateUpPercent")

    @rebate_up_percent.setter
    def rebate_up_percent(self, value):
        self._set_parameter("rebateUpPercent", value)
